from .avatar import Avatar
from .avatar_config import AvatarConfig
from .button import Button
from .button_config import ButtonConfig
from .checkbox import Checkbox
from .checkbox_config import CheckboxConfig
from .checkboxgroup import CheckboxGroup
from .checkboxgroup_config import CheckboxGroupConfig
from .colorpicker import Colorpicker
from .colorpicker_config import ColorpickerConfig
from .combo import Combo
from .combo_config import ComboConfig
from .container import Container
from .container_config import ContainerConfig
from .datepicker import Datepicker
from .datepicker_config import DatepickerConfig
from .fieldset import Fieldset
from .fieldset_config import FieldsetConfig
from .input import Input
from .input_config import InputConfig
from .radiogroup import RadioGroup
from .radiogroup_config import RadioGroupConfig
from .select import Select
from .select_config import SelectConfig
from .simplevault import SimpleVault
from .simplevault_config import SimpleVaultConfig
from .slider import Slider
from .slider_config import SliderConfig
from .spacer import Spacer
from .spacer_config import SpacerConfig
from .text import Text
from .text_config import TextConfig
from .textarea import Textarea
from .textarea_config import TextareaConfig
from .timepicker import Timepicker
from .timepicker_config import TimepickerConfig
from .toggle import Toggle
from .toggle_config import ToggleConfig
from .togglegroup import ToggleGroup
from .togglegroup_config import ToggleGroupConfig

__all__ = [
    "Avatar",
    "AvatarConfig",
    "Button",
    "ButtonConfig",
    "Checkbox",
    "CheckboxConfig",
    "CheckboxGroup",
    "CheckboxGroupConfig",
    "Colorpicker",
    "ColorpickerConfig",
    "Combo",
    "ComboConfig",
    "Container",
    "ContainerConfig",
    "Datepicker",
    "DatepickerConfig",
    "Fieldset",
    "FieldsetConfig",
    "Input",
    "InputConfig",
    "RadioGroup",
    "RadioGroupConfig",
    "Select",
    "SelectConfig",
    "SimpleVault",
    "SimpleVaultConfig",
    "Slider",
    "SliderConfig",
    "Spacer",
    "SpacerConfig",
    "Text",
    "TextConfig",
    "Textarea",
    "TextareaConfig",
    "Timepicker",
    "TimepickerConfig",
    "Toggle",
    "ToggleConfig",
    "ToggleGroup",
    "ToggleGroupConfig"   
]


from typing import Union, Dict, Any
from pyodide.ffi import create_proxy


class ButtonConfig:
    """
    Configuration class for the Button control.
    """
    def __init__(self,
                 name: str = None,
                 id: str = None,
                 text: str = None,
                 submit: bool = False,
                 url: str = None,
                 css: str = None,
                 disabled: bool = False,
                 height: Union[str, int] = "content",
                 hidden: bool = False,
                 padding: Union[str, int] = "5px",
                 width: Union[str, int] = "content",
                 circle: bool = False,
                 color: str = "primary",
                 full: bool = False,
                 icon: str = None,
                 loading: bool = False,
                 size: str = "medium",
                 view: str = "flat"):
        """
        :param name: (Optional) The name of the control.
        :param id: (Optional) The id of the control.
        :param text: (Optional) The text label of the button.
        :param submit: (Optional) Enables the button to send form data to a server.
        :param url: (Optional) The URL the post request with form data will be sent to.
        :param css: (Optional) Adds style classes to the control.
        :param disabled: (Optional) Whether the control is disabled.
        :param height: (Optional) The height of the control.
        :param hidden: (Optional) Whether the control is hidden.
        :param padding: (Optional) Sets padding between the cell and border.
        :param width: (Optional) The width of the control.
        :param circle: (Optional) Makes the corners of the button round.
        :param color: (Optional) Defines the color scheme of the button.
        :param full: (Optional) Extends the button to the full width of the form.
        :param icon: (Optional) The CSS class of an icon of the button.
        :param loading: (Optional) Adds a spinner into the button.
        :param size: (Optional) Defines the size of the button.
        :param view: (Optional) Defines the look of the button.
        """
        self.type = "button"
        self.name = name
        self.id = id
        self.text = text
        self.submit = submit
        self.url = url
        self.css = css
        self.disabled = disabled
        self.height = height
        self.hidden = hidden
        self.padding = padding
        self.width = width
        self.circle = circle
        self.color = color
        self.full = full
        self.icon = icon
        self.loading = loading
        self.size = size
        self.view = view

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the ButtonConfig into a dictionary format.
        """
        config_dict = {
            'type': self.type,
            'name': self.name,
            'id': self.id,
            'text': self.text,
            'submit': self.submit,
            'url': self.url,
            'css': self.css,
            'disabled': self.disabled,
            'height': self.height,
            'hidden': self.hidden,
            'padding': self.padding,
            'width': self.width,
            'circle': self.circle,
            'color': self.color,
            'full': self.full,
            'icon': self.icon,
            'loading': self.loading,
            'size': self.size,
            'view': self.view
        }
        # Remove None values
        return {k: v for k, v in config_dict.items() if v is not None}



from typing import List, Union, Dict, Any


def process_item(item):
    if isinstance(item, dict):
        if 'cols' in item:
            item['cols'] = [process_item(col) for col in item['cols']]
        if 'rows' in item:
            item['rows'] = [process_item(row) for row in item['rows']]
    elif hasattr(item, 'to_dict'):
        return item.to_dict()
    return item

class FormConfig:
    """
    Configuration class for Form.
    """
    def __init__(self,
                 align: str = "start",
                 cols: List[Dict[str, Any]] = None,
                 css: str = None,
                 disabled: bool = False,
                 height: Union[str, int] = "content",
                 hidden: bool = False,
                 padding: Union[str, int] = None,
                 rows: List[Dict[str, Any]] = None,
                 title: str = None,
                 width: Union[str, int] = "content"):
        """
        :param align: (Optional) Sets the alignment of controls inside the control group.
        :param cols: (Optional) Arranges controls inside the control group horizontally.
        :param css: (Optional) The name of a CSS class(es) applied to the control group.
        :param disabled: (Optional) Makes a form disabled.
        :param height: (Optional) Sets the height of the control group.
        :param hidden: (Optional) Defines whether a form is hidden.
        :param padding: (Optional) Sets padding for content inside the control group.
        :param rows: (Optional) Arranges controls inside the control group vertically.
        :param title: (Optional) Specifies the title of the control group.
        :param width: (Optional) Sets the width of the control group.
        """
        self.align = align
        self.cols = cols if cols else []
        self.css = css
        self.disabled = disabled
        self.height = height
        self.hidden = hidden
        self.padding = padding
        self.rows = rows if rows else []
        self.title = title
        self.width = width

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the FormConfig into a dictionary format that can be
        passed into the form constructor.
        """
        config_dict = {
            'align': self.align,
            'cols': self.cols,
            'css': self.css,
            'disabled': self.disabled,
            'height': self.height,
            'hidden': self.hidden,
            'padding': self.padding,
            'rows': self.rows,
            'title': self.title,
            'width': self.width
        }

        if 'cols' in config_dict:
            config_dict['cols'] = [process_item(col) for col in config_dict['cols']]
        if 'rows' in config_dict:
            config_dict['rows'] = [process_item(row) for row in config_dict['rows']]

            
        # Remove None values
        return {k: v for k, v in config_dict.items() if v is not None}




from typing import Any, Callable, TypeVar, List, Dict, Union

class CardFlowColumnConfig:
    """
    Configuration class for CardFlow columns.
    Similar to GridColumnConfig, lets you control properties like width, header text, etc.
    """
    def __init__(
        self,
        id: str,
        header: Union[str, Dict[str, Any]],
        width: Union[int, str] = "100px",
        align: str = None,
        hidden: bool = False,
        css: str = None,
        dataType: str = "str",
        dataFormat: str = "",
        applyFormat: bool = False,
        coltype: str = ""
    ):
        """
        Initializes a CardFlow column configuration.
        
        :param id: Unique identifier for the column.
        :param header: Header text or a more complex header config.
        :param width: Column width (e.g., "100px" or 100).
        :param align: Alignment for cell content.
        :param hidden: Whether the column is hidden.
        :param css: Additional CSS classes.
        :param dataType: Type of the data (e.g., "str", "int", "time").
        :param dataFormat: Used if dataType is "time" (e.g. "HH:MM AM/PM").
        :param applyFormat: Whether to apply the dataFormat for display.
        """
        self.id = id
        self.header = header
        self.width = width
        self.align = align
        self.hidden = hidden
        self.css = css
        self.dataType = dataType
        self.dataFormat = dataFormat
        self.applyFormat = applyFormat
        self.coltype = coltype

    def to_dict(self) -> Dict[str, Any]:
        config_dict = {
            "id": self.id,
            "header": self.header,
            "width": self.width,
            "align": self.align,
            "hidden": self.hidden,
            "css": self.css,
            "dataType": self.dataType,
            "dataFormat": self.dataFormat,
            "applyFormat": self.applyFormat,
            "type": self.coltype
        }
        # Remove keys that are None so we don't pass unnecessary data to JS
        return {k: v for k, v in config_dict.items() if v is not None}


class CardFlowConfig:
    """
    Configuration class for the CardFlow widget.
    Example structure:
    {
        "columns": [
            { "id": "name", "header": "Name:" },
            { "id": "vehicle", "header": "Vehicle:" },
            { "id": "status", "header": "Status:" },
            { "id": "promise_time", "header": "Promise Time:" }
        ],
        "data": [ ... ],
        "cardHeight": "120px",
        "stacked": True,
        "defaultExpandedHeight": "300px"
    }
    """
    def __init__(
        self,
        columns: List[Union[Dict[str, Any], CardFlowColumnConfig]] = None,
        data: List[Dict[str, Any]] = None,
        editable: bool = True,
        group: Dict[str, Any] = None,
        groupable: bool = True,
        hideExpandCollapse: bool = False,
        autoCollapse: bool = False,
        optionItems: List[Dict[str, Any]] = None,
        sortDisabled: bool = False,
        showHeader: bool = True,
        showSort: bool = True,
        sortHeader: str = "",
        showDataHeaders: bool = True,
        showOptions: bool = True,
        fontSize: str = "",
        toolbar_font_family: str = "",
        cardHeight: str = None,  # Height of collapsed cards
        stacked: bool = False,
        defaultExpandedHeight: str = "300px",  # NEW: Default height for expanded cards
        gpu: bool = False,
        gpu_widget_id: str = "",
    ):
        """
        Initializes the CardFlowConfig.

        :param columns: List of column configurations (dict or CardFlowColumnConfig).
        :param data: A list of dictionaries containing the CardFlow data (each representing one card).
        :param editable: Whether the CardFlow is editable.
        :param group: Optional grouping config (not always used).
        :param groupable: Whether grouping is enabled (not always used).
        :param hideExpandCollapse: If set to True, you might hide the expand/collapse UI.
        :param optionItems: Specifies items for the "options" dropdown.
        :param sortDisabled: Disables the sorting toolbar if True.
        :param sortHeader: Label/text to show on the sorting toolbar.
        :param cardHeight: Height (CSS) of each card row, e.g. "100px". Defaults to "auto" in JS if not set.
        :param stacked: If True, card data displays in vertical (stacked) format rather than side-by-side grid.
        :param defaultExpandedHeight: Default height for expanded card content areas (e.g., "300px").
        """
        self.columns = columns if columns is not None else []
        self.data = data if data is not None else []
        self.editable = editable
        self.group = group if group is not None else {}
        self.groupable = groupable
        self.hideExpandCollapse = hideExpandCollapse
        self.autoCollapse = autoCollapse
        self.sortDisabled = sortDisabled
        self.sortHeader = sortHeader
        self.cardHeight = cardHeight
        self.stacked = stacked
        self.defaultExpandedHeight = defaultExpandedHeight
        self.optionItems = optionItems
        self.showHeader = showHeader
        self.showSort = showSort
        self.showDataHeaders = showDataHeaders
        self.fontSize = fontSize
        self.toolbar_font_family = toolbar_font_family
        self.showOptions = showOptions
        self.gpu = gpu
        self.gpu_widget_id = gpu_widget_id

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the CardFlowConfig into a dictionary format for the JS code.
        """
        config_dict = {
            "columns": [
                col.to_dict() if hasattr(col, "to_dict") else col 
                for col in self.columns
            ],
            "data": self.data,
            "editable": self.editable,
            "group": self.group,
            "groupable": self.groupable,
            "hideExpandCollapse": self.hideExpandCollapse,
            "autoCollapse": self.autoCollapse,
            "optionItems": self.optionItems,
            "sortDisabled": self.sortDisabled,
            "sortHeader": self.sortHeader,
            # existing parameters
            "cardHeight": self.cardHeight,
            "stacked": self.stacked,
            "showHeader": self.showHeader,
            "showSort": self.showSort,
            "showDataHeaders": self.showDataHeaders,
            "fontSize": self.fontSize,
            "toolbarFontFamily": self.toolbar_font_family,
            "showOptions": self.showOptions,
            # new parameter
            "defaultExpandedHeight": self.defaultExpandedHeight,
            "gpu": self.gpu,
            "gpuWidgetId": self.gpu_widget_id or None,
        }
        # Remove keys with None values
        return {k: v for k, v in config_dict.items() if v is not None}



"""
pyTincture Form widget implementation
"""

from typing import Any, Callable, Dict, List, Union
import json
from pyodide.ffi import create_proxy
import js

from .form_config import FormConfig


class Form:
    def __init__(self, config: FormConfig = None, widget_parent: Any = None):
        """Initializes the Form instance."""
        if config is None:
            config = FormConfig()
        config_dict = config.to_dict()
        self.form = js.dhx.Form.new(widget_parent, js.JSON.parse(json.dumps(config_dict)))

    """ Form API Functions """

    def blur(self, name: str = None) -> None:
        """Removes focus from a control of the Form."""
        self.form.blur(name)

    def clear(self, method: str = None) -> None:
        """Clears the form."""
        self.form.clear(method)

    def destructor(self) -> None:
        """Removes the Form instance and releases occupied resources."""
        self.form.destructor()

    def disable(self) -> None:
        """Disables the Form."""
        self.form.disable()

    def enable(self) -> None:
        """Enables the Form."""
        self.form.enable()

    def for_each(self, callback: Callable[[Any, int, List[Any]], Any]) -> None:
        """Iterates over all controls of the Form."""
        proxy_callback = create_proxy(callback)
        self.form.forEach(proxy_callback)

    def get_item(self, name: str) -> Any:
        """Gives access to the object of a Form control."""
        return self.form.getItem(name)

    def get_properties(self, name: str = None) -> Union[Dict[str, Any], Dict[str, Dict[str, Any]]]:
        """Returns objects with available configuration attributes of Form controls."""
        return self.form.getProperties(name)

    def get_value(self, as_form_data: bool = False) -> Dict[str, Any]:
        """Gets current values/states of controls."""
        return self.form.getValue(as_form_data)

    def hide(self) -> None:
        """Hides the Form."""
        self.form.hide()

    def is_disabled(self, name: str = None) -> bool:
        """Checks whether the Form or a control is disabled."""
        return self.form.isDisabled(name)

    def is_visible(self, name: str = None) -> bool:
        """Checks whether the Form or a control is visible."""
        return self.form.isVisible(name)

    def paint(self) -> None:
        """Repaints the Form on the page."""
        self.form.paint()

    def send(self, url: str, method: str = "POST", as_form_data: bool = False) -> Any:
        """Sends the Form to the server."""
        return self.form.send(url, method, as_form_data)

    def set_focus(self, name: str) -> None:
        """Sets focus to a Form control by its name or id."""
        self.form.setFocus(name)

    def set_properties(self, arg: Union[str, Dict[str, Dict[str, Any]]], properties: Dict[str, Any] = None) -> None:
        """Allows changing available configuration attributes of Form controls dynamically."""
        if isinstance(arg, str):
            self.form.setProperties(arg, js.JSON.parse(json.dumps(properties)))
        elif isinstance(arg, dict):
            props = {k: js.JSON.parse(json.dumps(v)) for k, v in arg.items()}
            self.form.setProperties(props)
        else:
            raise TypeError("Argument must be a string or a dictionary")

    def set_value(self, obj: Dict[str, Any]) -> None:
        """Sets values/states for controls."""
        self.form.setValue(js.JSON.parse(json.dumps(obj)))

    def show(self) -> None:
        """Shows the Form on the page."""
        self.form.show()

    def validate(self, silent: bool = False) -> bool:
        """Validates form fields."""
        return self.form.validate(silent)

    """ Form Events """

    def add_event_handler(self, event_name: str, handler: Callable) -> None:
        """Helper to add event handlers dynamically."""
        event_proxy = create_proxy(handler)
        self.form.events.on(event_name, event_proxy)

    def on_after_change_properties(self, handler: Callable[[str, Dict[str, Any]], None]) -> None:
        """Fires after configuration attributes of a Form control have been changed dynamically."""
        self.add_event_handler('afterChangeProperties', handler)

    def on_after_hide(self, handler: Callable[[str, Any, str], None]) -> None:
        """Fires after a Form control or its element is hidden."""
        self.add_event_handler('afterHide', handler)

    def on_after_send(self, handler: Callable[[], None]) -> None:
        """Fires after a form is sent to the server."""
        self.add_event_handler('afterSend', handler)

    def on_after_show(self, handler: Callable[[str, Any, str], None]) -> None:
        """Fires after a Form control or its element is shown."""
        self.add_event_handler('afterShow', handler)

    def on_after_validate(self, handler: Callable[[str, Any, bool], None]) -> None:
        """Fires after validation of form fields is finished."""
        self.add_event_handler('afterValidate', handler)

    def on_before_change(self, handler: Callable[[str, Any], Union[bool, None]]) -> None:
        """Fires before changing the value of a control."""
        def event_handler(name, value):
            result = handler(name, value)
            if result is False:
                return js.Boolean(False)
        self.form.events.on('beforeChange', create_proxy(event_handler))

    def on_before_change_properties(self, handler: Callable[[str, Dict[str, Any]], Union[bool, None]]) -> None:
        """Fires before configuration attributes of a Form control are changed dynamically."""
        def event_handler(name, properties):
            result = handler(name, properties)
            if result is False:
                return js.Boolean(False)
        self.form.events.on('beforeChangeProperties', create_proxy(event_handler))

    def on_before_hide(self, handler: Callable[[Union[str, int], Any, str], Union[bool, None]]) -> None:
        """Fires before a Form control or its element is hidden."""
        def event_handler(name, value=None, id=None):
            result = handler(name, value, id)
            if result is False:
                return js.Boolean(False)
        self.form.events.on('beforeHide', create_proxy(event_handler))

    def on_before_send(self, handler: Callable[[], Union[bool, None]]) -> None:
        """Fires before a form is sent to the server."""
        def event_handler():
            result = handler()
            if result is False:
                return js.Boolean(False)
        self.form.events.on('beforeSend', create_proxy(event_handler))

    def on_before_show(self, handler: Callable[[str, Any, str], Union[bool, None]]) -> None:
        """Fires before a Form control or its element is shown."""
        def event_handler(name, value=None, id=None):
            result = handler(name, value, id)
            if result is False:
                return js.Boolean(False)
        self.form.events.on('beforeShow', create_proxy(event_handler))

    def on_before_validate(self, handler: Callable[[str, Any], Union[bool, None]]) -> None:
        """Fires before validation of form fields has started."""
        def event_handler(name, value):
            result = handler(name, value)
            if result is False:
                return js.Boolean(False)
        self.form.events.on('beforeValidate', create_proxy(event_handler))

    def on_blur(self, handler: Callable[[str, Any, str], None]) -> None:
        """Fires when a control of Form has lost focus."""
        self.add_event_handler('blur', handler)

    def on_change(self, handler: Callable[[str, Any], None]) -> None:
        """Fires on changing the value of a control."""
        self.add_event_handler('change', handler)

    def on_click(self, handler: Callable[[str, Any], Any]) -> None:
        """Fires after a click on a button in a form."""
        self.add_event_handler('click', handler)

    def on_focus(self, handler: Callable[[str, Any, str], None]) -> None:
        """Fires when a control of Form has received focus."""
        self.add_event_handler('focus', handler)

    def on_keydown(self, handler: Callable[[Any, str, str], None]) -> None:
        """Fires when any key is pressed."""
        self.add_event_handler('keydown', handler)

    """ Form Properties """

    @property
    def align(self) -> str:
        """Gets or sets the alignment of controls inside the control group."""
        return self.form.config.align

    @align.setter
    def align(self, value: str) -> None:
        self.form.config.align = value

    @property
    def cols(self) -> List[Dict[str, Any]]:
        """Gets or sets the columns of controls inside the control group."""
        return self.form.config.cols

    @cols.setter
    def cols(self, value: List[Dict[str, Any]]) -> None:
        self.form.config.cols = value

    @property
    def css(self) -> str:
        """Gets or sets the CSS classes applied to the control group."""
        return self.form.config.css

    @css.setter
    def css(self, value: str) -> None:
        self.form.config.css = value

    @property
    def disabled(self) -> bool:
        """Gets or sets whether the Form is disabled."""
        return self.form.config.disabled

    @disabled.setter
    def disabled(self, value: bool) -> None:
        self.form.config.disabled = value

    @property
    def height(self) -> Union[str, int]:
        """Gets or sets the height of the control group."""
        return self.form.config.height

    @height.setter
    def height(self, value: Union[str, int, str]) -> None:
        self.form.config.height = value

    @property
    def hidden(self) -> bool:
        """Gets or sets whether the Form is hidden."""
        return self.form.config.hidden

    @hidden.setter
    def hidden(self, value: bool) -> None:
        self.form.config.hidden = value

    @property
    def padding(self) -> Union[str, int]:
        """Gets or sets the padding for content inside the control group."""
        return self.form.config.padding

    @padding.setter
    def padding(self, value: Union[str, int]) -> None:
        self.form.config.padding = value

    @property
    def rows(self) -> List[Dict[str, Any]]:
        """Gets or sets the rows of controls inside the control group."""
        return self.form.config.rows

    @rows.setter
    def rows(self, value: List[Dict[str, Any]]) -> None:
        self.form.config.rows = value

    @property
    def title(self) -> str:
        """Gets or sets the title of the control group."""
        return self.form.config.title

    @title.setter
    def title(self, value: str) -> None:
        self.form.config.title = value

    @property
    def width(self) -> Union[str, int]:
        """Gets or sets the width of the control group."""
        return self.form.config.width

    @width.setter
    def width(self, value: Union[str, int, str]) -> None:
        self.form.config.width = value




from typing import Any, Callable, Dict, Union
from pyodide.ffi import create_proxy


class ListboxConfig:
    """
    Configuration class for the ListBox widget.
    """
    def __init__(self,
                 data: list = None,
                 css: str = None,
                 dragCopy: bool = False,
                 dragMode: str = None,
                 editable: bool = False,
                 eventHandlers: Dict[str, Dict[str, Callable]] = None,
                 height: Union[int, str] = "auto",
                 htmlEnable: bool = True,
                 itemHeight: Union[int, str] = 37,
                 keyNavigation: Union[bool, Callable[[], bool]] = True,
                 multiselection: Union[bool, str] = False,
                 selection: bool = True,
                 template: Callable[[Dict[str, Any]], str] = None,
                 virtual: bool = False):
        """
        Initializes the ListBoxConfig.

        :param data: (Optional) List of data items.
        :param css: (Optional) Adds a CSS class(es) to the component.
        :param dragCopy: (Optional) Defines that an item is copied to a target during drag-n-drop.
        :param dragMode: (Optional) Enables drag-n-drop in ListBox.
        :param editable: (Optional) Enables editing in ListBox.
        :param eventHandlers: (Optional) Adds event handlers to HTML elements of a custom template.
        :param height: (Optional) Sets the height of ListBox.
        :param htmlEnable: (Optional) Enables/disables rendering of HTML content.
        :param itemHeight: (Optional) Sets the height of an item.
        :param keyNavigation: (Optional) Enables/disables navigation by arrow keys.
        :param multiselection: (Optional) Enables multiselection mode.
        :param selection: (Optional) Enables selection of ListBox items.
        :param template: (Optional) Specifies a template for ListBox items.
        :param virtual: (Optional) Enables dynamic rendering of ListBox items.
        """
        self.data = data
        self.css = css
        self.dragCopy = dragCopy
        self.dragMode = dragMode
        self.editable = editable
        self.eventHandlers = eventHandlers
        self.height = height
        self.htmlEnable = htmlEnable
        self.itemHeight = itemHeight
        self.keyNavigation = keyNavigation
        self.multiselection = multiselection
        self.selection = selection
        self.template = template
        self.virtual = virtual

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the ListBoxConfig into a dictionary format.
        """
        config_dict = {
            'data': self.data,
            'css': self.css,
            'dragCopy': self.dragCopy,
            'dragMode': self.dragMode,
            'editable': self.editable,
            'eventHandlers': self.eventHandlers,
            'height': self.height,
            'htmlEnable': self.htmlEnable,
            'itemHeight': self.itemHeight,
            'keyNavigation': self.keyNavigation,
            'multiselection': self.multiselection,
            'selection': self.selection,
            'template': self.template,
            'virtual': self.virtual,
        }
        # Remove None values
        config_dict = {k: v for k, v in config_dict.items() if v is not None}

        # Handle functions (e.g., template, keyNavigation)
        if 'template' in config_dict and callable(config_dict['template']):
            # Assuming template is a JavaScript function or needs to be converted
            config_dict['template'] = create_proxy(self.template)

        if 'keyNavigation' in config_dict and callable(config_dict['keyNavigation']):
            config_dict['keyNavigation'] = create_proxy(self.keyNavigation)

        return config_dict



"""Demo harness for the CardFlow widget (vehicle repair work orders)."""

import os
import sys
from typing import Dict, List, Optional

from dhxpyt.layout import CellConfig, LayoutConfig, MainWindow
from dhxpyt.cardflow import CardFlow, CardFlowConfig, CardFlowColumnConfig


class CardFlowDemo(MainWindow):
    layout_config = LayoutConfig(
        type="line",
        rows=[CellConfig(id="flow", height="100%")],
    )

    def __init__(self) -> None:
        super().__init__()
        self.set_theme(os.getenv("CARDFLOW_APP_THEME", "dark"))
        self._cardflow: Optional[CardFlow] = None
        self._orders: List[Dict[str, str]] = []

    def load_ui(self) -> None:  # type: ignore[override]
        config = self._build_config()
        self._orders = config.data or []
        self._cardflow = self.add_cardflow("flow", cardflow_config=config)
        if not self._cardflow:
            return

        self._cardflow.on_card_expand(self._on_card_expand)
        self._cardflow.on_card_collapse(self._log_event("cardCollapse"))
        self._cardflow.on_options(self._log_event("cardOptions"))

    # ------------------------------------------------------------------
    # Configuration helpers
    # ------------------------------------------------------------------

    def _build_config(self) -> CardFlowConfig:
        columns = [
            CardFlowColumnConfig(id="ticket", header="Ticket #", width="130px"),
            CardFlowColumnConfig(id="vehicle", header="Vehicle", width="220px"),
            CardFlowColumnConfig(id="technician", header="Technician", width="170px"),
            CardFlowColumnConfig(
                id="status",
                header="Status",
                width="150px",
            ),
            CardFlowColumnConfig(
                id="promise_time",
                header="Promise",
                width="140px",
                dataType="time",
                dataFormat="HH:MM AM/PM",
                applyFormat=True,
            ),
            CardFlowColumnConfig(id="bay", header="Bay", width="110px"),
            {"type": "stretch"},
        ]

        data = [
            {
                "id": "order-101",
                "ticket": "RO-101",
                "vehicle": "2019 Ford F-150",
                "technician": "Alex Rivera",
                "status": "Diagnostics",
                "promise_time": "09:30 AM",
                "bay": "Bay 2",
                "_color": "#2563eb",
            },
            {
                "id": "order-102",
                "ticket": "RO-102",
                "vehicle": "2012 Nissan Altima",
                "technician": "Alex Rivera",
                "status": "Waiting Parts",
                "promise_time": "11:45 AM",
                "bay": "Bay 5",
            },
            {
                "id": "order-103",
                "ticket": "RO-103",
                "vehicle": "2021 Honda Civic",
                "technician": "Bailey Kim",
                "status": "Intake",
                "promise_time": "12:15 PM",
                "bay": "Staging",
                "_color": "#d97706",
            },
            {
                "id": "order-104",
                "ticket": "RO-104",
                "vehicle": "2020 Tesla Model 3",
                "technician": "Bailey Kim",
                "status": "Repair",
                "promise_time": "02:05 PM",
                "bay": "Bay 4",
            },
            {
                "id": "order-105",
                "ticket": "RO-105",
                "vehicle": "2017 BMW X5",
                "technician": "Marco Diaz",
                "status": "Quality Check",
                "promise_time": "03:40 PM",
                "bay": "Bay 7",
                "_color": "#0891b2",
            },
        ]

        option_items = [
            {"id": "details", "value": "View Details"},
            {"id": "reassign", "value": "Reassign"},
            {"id": "hold", "value": "Place on Hold"},
            {"id": "cancel", "value": "Cancel Ticket"},
        ]

        return CardFlowConfig(
            columns=columns,
            data=data,
            sortHeader="Repair Workflow",
            optionItems=option_items,
            showHeader=True,
            showSort=True,
            showDataHeaders=True,
            fontSize="13px",
            defaultExpandedHeight="280px",
        )

    # ------------------------------------------------------------------
    # Event helpers
    # ------------------------------------------------------------------

    def _on_card_expand(self, card_id: str, *_args) -> None:
        print(f"[cardflow_app] cardExpand: {card_id}")
        if not self._cardflow:
            return
        order = next((item for item in self._orders if item["id"] == card_id), None)
        if not order:
            return
        accent = order.get("_color") or "#2563eb"
        self._cardflow.set_row_color(card_id, accent)

    @staticmethod
    def _log_event(name: str):
        def handler(*args, **kwargs):
            payload = args[0] if args else None
            print(f"[cardflow_app] {name}: {payload}")
        return handler


if __name__ == "__main__" and sys.platform != "emscripten":
    from pytincture import launch_service

    launch_service(modules_folder=".")



"""
Slider widget implementation
"""

from typing import Any, Callable, Dict, List, Optional, Union
import json
from pyodide.ffi import create_proxy
import js

from .slider_config import SliderConfig


class Slider:
    def __init__(self, config: SliderConfig, widget_parent: str = None):
        """
        Initializes the Slider widget.

        :param config: The SliderConfig object containing the slider configuration.
        :param widget_parent: (Optional) The ID of the HTML element where the slider will be attached.
        """
        config_dict = config.to_dict()

        # Handle the tickTemplate function separately
        tick_template = None
        if 'tickTemplate' in config_dict:
            tick_template = config_dict.pop('tickTemplate')
            # Create a proxy for the JavaScript side
            config_dict['tickTemplate'] = create_proxy(tick_template)

        # Create the Slider instance
        self.slider = js.dhx.Slider.new(widget_parent, js.JSON.parse(json.dumps(config_dict)))
        if tick_template:
            # Assign the tickTemplate function
            self.slider.config.tickTemplate = config_dict['tickTemplate']

    """ Slider API Functions """

    def blur(self) -> None:
        """Removes focus from a thumb of Slider."""
        self.slider.blur()

    def destructor(self) -> None:
        """Destroys the slider instance and releases occupied resources."""
        self.slider.destructor()

    def disable(self) -> None:
        """Disables the slider."""
        self.slider.disable()

    def enable(self) -> None:
        """Enables the slider."""
        self.slider.enable()

    def focus(self, extra: bool = None) -> None:
        """
        Sets focus to a thumb of Slider.

        :param extra: (Optional) If the range mode is activated, True will set focus to the second thumb.
        """
        if extra is not None:
            self.slider.focus(extra)
        else:
            self.slider.focus()

    def get_value(self) -> List[float]:
        """
        Returns the current value of Slider.

        :return: An array with the current value of the slider.
        """
        value = self.slider.getValue()
        return value.to_py()

    def is_disabled(self) -> bool:
        """
        Checks whether Slider is disabled.

        :return: True if the slider is disabled; otherwise, False.
        """
        return self.slider.isDisabled()

    def paint(self) -> None:
        """Repaints Slider on a page."""
        self.slider.paint()

    def set_value(self, value: Union[str, float, List[float]]) -> None:
        """
        Sets a value for the slider.

        :param value: The value to be set for Slider.
        """
        self.slider.setValue(value)

    """ Slider Event Handlers """

    def add_event_handler(self, event_name: str, handler: Callable) -> None:
        """
        Adds an event handler for the specified event.

        :param event_name: The name of the event.
        :param handler: The handler function to attach.
        """
        event_proxy = create_proxy(handler)
        self.slider.events.on(event_name, event_proxy)

    def on_before_change(self, handler: Callable[[float, float, bool], Union[bool, None]]) -> None:
        """
        Fires before changing of the slider value.

        :param handler: The handler function with parameters value (float), oldValue (float), isRange (bool).
                        Return False to prevent changing the slider value.
        """
        def event_handler(value, oldValue, isRange):
            result = handler(value, oldValue, isRange)
            if result is False:
                return js.Boolean(False)
        self.slider.events.on('beforeChange', create_proxy(event_handler))

    def on_blur(self, handler: Callable[[], None]) -> None:
        """
        Fires when a thumb of Slider has lost focus.

        :param handler: The handler function.
        """
        self.add_event_handler('blur', handler)

    def on_change(self, handler: Callable[[float, float, bool], None]) -> None:
        """
        Fires on change of the slider value.

        :param handler: The handler function with parameters value (float), oldValue (float), isRange (bool).
        """
        def event_handler(value, oldValue, isRange):
            handler(value, oldValue, isRange)
        self.slider.events.on('change', create_proxy(event_handler))

    def on_focus(self, handler: Callable[[], None]) -> None:
        """
        Fires when a thumb of Slider has received focus.

        :param handler: The handler function.
        """
        self.add_event_handler('focus', handler)

    def on_keydown(self, handler: Callable[[Any], None]) -> None:
        """
        Fires when any key is pressed and a thumb of Slider is in focus.

        :param handler: The handler function with parameter event (KeyboardEvent).
        """
        def event_handler(event):
            handler(event)
        self.slider.events.on('keydown', create_proxy(event_handler))

    def on_mousedown(self, handler: Callable[[Any], None]) -> None:
        """
        Fires on pressing the left mouse button over the slider thumb.

        :param handler: The handler function with parameter event (Event).
        """
        def event_handler(event):
            handler(event)
        self.slider.events.on('mousedown', create_proxy(event_handler))

    def on_mouseup(self, handler: Callable[[Any], None]) -> None:
        """
        Fires on releasing the left mouse button over the slider thumb.

        :param handler: The handler function with parameter event (Event).
        """
        def event_handler(event):
            handler(event)
        self.slider.events.on('mouseUp', create_proxy(event_handler))




"""
Pagination widget implementation
"""

from typing import Any, Callable
import json
from pyodide.ffi import create_proxy
import js

from .pagination_config import PaginationConfig


class Pagination:
    def __init__(self, config: PaginationConfig = None, widget_parent: str = None):
        """
        Initializes the Pagination widget.

        :param config: (Required) The PaginationConfig object containing the pagination configuration.
        :param widget_parent: (Optional) The ID of the HTML element where the pagination will be attached.
        """
        if config is None:
            raise ValueError("PaginationConfig is required for initializing Pagination widget.")
        config_dict = config.to_dict()
        # Assuming that `config.data` is a DHTMLX DataCollection instance
        # We need to handle data initialization appropriately
        # For now, we will pass the data collection directly

        # Create the Pagination instance
        self.pagination = js.dhx.Pagination.new(widget_parent, js.JSON.parse(json.dumps(config_dict)))

    """ Pagination API Functions """

    def destructor(self) -> None:
        """Destroys the Pagination instance and releases resources."""
        self.pagination.destructor()

    def get_page(self) -> int:
        """Returns the index of the active page."""
        return self.pagination.getPage()

    def get_pages_count(self) -> int:
        """Gets the total number of pages in the pagination."""
        return self.pagination.getPagesCount()

    def get_page_size(self) -> int:
        """Returns the number of items displayed per page."""
        return self.pagination.getPageSize()

    def set_page(self, page: int) -> None:
        """Sets an active page in the related widget."""
        self.pagination.setPage(page)

    def set_page_size(self, size: int) -> None:
        """Sets the number of items displayed on a page."""
        self.pagination.setPageSize(size)

    """ Pagination Event Handlers """

    def add_event_handler(self, event_name: str, handler: Callable) -> None:
        """Adds an event handler for the specified event."""
        event_proxy = create_proxy(handler)
        self.pagination.events.on(event_name, event_proxy)

    def on_change(self, handler: Callable[[int, int], None]) -> None:
        """Fires on changing the active page."""
        def event_handler(index, previousIndex):
            handler(index, previousIndex)
        self.pagination.events.on('change', create_proxy(event_handler))



"""
Sidebar widget implementation
"""

from typing import Any, Callable, Dict, List, Optional, Union
import json
from pyodide.ffi import create_proxy
import js

from .sidebar_config import SidebarConfig


class Sidebar:
    def __init__(self, config: SidebarConfig = None, widget_parent: str = None):
        """
        Initializes the Sidebar widget.

        :param config: (Optional) The SidebarConfig object containing the sidebar configuration.
        :param widget_parent: (Optional) The ID of the HTML element where the sidebar will be attached.
        """
        if config is None:
            config = SidebarConfig()
        config_dict = config.to_dict()
        # Create the Sidebar instance
        self.sidebar = js.dhx.Sidebar.new(widget_parent, js.JSON.parse(json.dumps(config_dict)))

    """ Sidebar API Functions """

    def collapse(self) -> None:
        """Collapses the sidebar."""
        self.sidebar.collapse()

    def destructor(self) -> None:
        """Destroys the Sidebar instance and releases resources."""
        self.sidebar.destructor()

    def disable(self, ids: Union[str, int, List[Union[str, int]]] = None) -> None:
        """Disables and dims items of Sidebar."""
        self.sidebar.disable(ids)

    def enable(self, ids: Union[str, int, List[Union[str, int]]] = None) -> None:
        """Enables disabled items of Sidebar."""
        self.sidebar.enable(ids)

    def expand(self) -> None:
        """Expands the sidebar."""
        self.sidebar.expand()

    def get_selected(self) -> List[Union[str, int]]:
        """Returns an array of IDs of selected items."""
        selected = self.sidebar.getSelected()
        return [item for item in selected.to_py()]

    def hide(self, ids: Union[str, int, List[Union[str, int]]] = None) -> None:
        """Hides items of Sidebar."""
        self.sidebar.hide(ids)

    def is_collapsed(self) -> bool:
        """Checks whether Sidebar is collapsed."""
        return self.sidebar.isCollapsed()

    def is_disabled(self, id: Union[str, int]) -> bool:
        """Checks whether an item of Sidebar is disabled."""
        return self.sidebar.isDisabled(id)

    def is_selected(self, id: Union[str, int]) -> bool:
        """Checks whether a specified Sidebar item is selected."""
        return self.sidebar.isSelected(id)

    def paint(self) -> None:
        """Repaints Sidebar on a page."""
        self.sidebar.paint()

    def select(self, id: Union[str, int], unselect: bool = True) -> None:
        """Selects a specified Sidebar item."""
        self.sidebar.select(id, unselect)

    def show(self, ids: Union[str, int, List[Union[str, int]]] = None) -> None:
        """Shows items of Sidebar."""
        self.sidebar.show(ids)

    def toggle(self) -> None:
        """Expands/collapses Sidebar."""
        self.sidebar.toggle()

    def unselect(self, id: Union[str, int] = None) -> None:
        """Unselects a selected Sidebar item."""
        self.sidebar.unselect(id)

    """ Sidebar Event Handlers """

    def add_event_handler(self, event_name: str, handler: Callable) -> None:
        """
        Adds an event handler for the specified event.

        :param event_name: The name of the event.
        :param handler: The handler function to attach.
        """
        event_proxy = create_proxy(handler)
        self.sidebar.events.on(event_name, event_proxy)

    def on_after_collapse(self, handler: Callable[[], None]) -> None:
        """Fires after collapsing a sidebar."""
        self.add_event_handler('afterCollapse', handler)

    def on_after_expand(self, handler: Callable[[], None]) -> None:
        """Fires after expanding a sidebar."""
        self.add_event_handler('afterExpand', handler)

    def on_after_hide(self, handler: Callable[[Any], None]) -> None:
        """
        Fires after hiding a sub-item of Sidebar.

        :param handler: The handler function with parameter events (Event).
        """
        self.add_event_handler('afterHide', handler)

    def on_before_collapse(self, handler: Callable[[], Union[bool, None]]) -> None:
        """
        Fires before collapsing a sidebar.

        :param handler: The handler function that returns False to prevent collapsing.
        """
        def event_handler():
            result = handler()
            if result is False:
                return js.Boolean(False)
        self.sidebar.events.on('beforeCollapse', create_proxy(event_handler))

    def on_before_expand(self, handler: Callable[[], Union[bool, None]]) -> None:
        """
        Fires before expanding a sidebar.

        :param handler: The handler function that returns False to prevent expanding.
        """
        def event_handler():
            result = handler()
            if result is False:
                return js.Boolean(False)
        self.sidebar.events.on('beforeExpand', create_proxy(event_handler))

    def on_before_hide(self, handler: Callable[[Union[str, int], Any], Union[bool, None]]) -> None:
        """
        Fires before hiding a sub-item of Sidebar.

        :param handler: The handler function with parameters id (str or int), events (Event).
        """
        def event_handler(id, events):
            result = handler(id, events)
            if result is False:
                return js.Boolean(False)
        self.sidebar.events.on('beforeHide', create_proxy(event_handler))

    def on_click(self, handler: Callable[[Union[str, int], Any], None]) -> None:
        """
        Fires after a click on a control.

        :param handler: The handler function with parameters id (str or int), events (Event).
        """
        def event_handler(id, events):
            handler(id, events)
        self.sidebar.events.on('click', create_proxy(event_handler))

    def on_input_blur(self, handler: Callable[[Union[str, int]], None]) -> None:
        """
        Fires when a control is blurred.

        :param handler: The handler function with parameter id (str or int).
        """
        self.add_event_handler('inputBlur', handler)

    def on_input_created(self, handler: Callable[[Union[str, int], Any], None]) -> None:
        """
        Fires when a new input is added.

        :param handler: The handler function with parameters id (str or int), input (HTMLInputElement).
        """
        def event_handler(id, input):
            handler(id, input)
        self.sidebar.events.on('inputCreated', create_proxy(event_handler))

    def on_input_focus(self, handler: Callable[[Union[str, int]], None]) -> None:
        """
        Fires when a control is focused.

        :param handler: The handler function with parameter id (str or int).
        """
        self.add_event_handler('inputFocus', handler)

    def on_keydown(self, handler: Callable[[Any, Optional[str]], None]) -> None:
        """
        Fires when any key is pressed and a Sidebar option is in focus.

        :param handler: The handler function with parameters event (KeyboardEvent), id (str or None).
        """
        def event_handler(event, id=None):
            handler(event, id)
        self.sidebar.events.on('keydown', create_proxy(event_handler))

    def on_open_menu(self, handler: Callable[[Union[str, int]], None]) -> None:
        """
        Fires on expanding a menu control.

        :param handler: The handler function with parameter id (str or int).
        """
        self.add_event_handler('openMenu', handler)



"""
RadioGroup control implementation for the Form widget
"""

from typing import Any, Callable, Dict, Union, List
import json
from pyodide.ffi import create_proxy
import js

from .radiogroup_config import RadioGroupConfig


class RadioGroup:
    def __init__(self, config: RadioGroupConfig = None, widget_parent: Any = None):
        """Initializes the RadioGroup control."""
        if config is None:
            config = RadioGroupConfig()
        config_dict = config.to_dict()
        print(config_dict)
        self.radiogroup = js.dhx.FormControl.new(widget_parent, js.JSON.parse(json.dumps(config_dict)))

    """ RadioGroup API Functions """

    def blur(self) -> None:
        """Removes focus from the RadioGroup control."""
        self.radiogroup.blur()

    def clear(self) -> None:
        """Clears the value of the RadioGroup control."""
        self.radiogroup.clear()

    def clear_validate(self) -> None:
        """Clears validation of the RadioGroup control."""
        self.radiogroup.clearValidate()

    def destructor(self) -> None:
        """Removes the RadioGroup instance and releases the occupied resources."""
        self.radiogroup.destructor()

    def disable(self, id: str = None) -> None:
        """Disables the RadioGroup control or a specific element inside the control."""
        self.radiogroup.disable(id)

    def enable(self, id: str = None) -> None:
        """Enables the RadioGroup control or a specific element inside the control."""
        self.radiogroup.enable(id)

    def focus(self, id: str = None) -> None:
        """Sets focus to the radio button of the RadioGroup control by its id."""
        self.radiogroup.focus(id)

    def get_properties(self, id: str = None) -> Dict[str, Any]:
        """Returns the configuration attributes of the control or a specific radio button."""
        props = self.radiogroup.getProperties(id)
        return props.to_py() if props else {}

    def get_value(self) -> str:
        """Returns the current value of the RadioGroup control."""
        return self.radiogroup.getValue()

    def hide(self, id: str = None) -> None:
        """Hides either a radio button of RadioGroup or the whole RadioGroup."""
        self.radiogroup.hide(id)

    def is_disabled(self, id: str = None) -> bool:
        """Checks whether the RadioGroup control or a specific element is disabled."""
        return self.radiogroup.isDisabled(id)

    def is_visible(self, id: str = None) -> bool:
        """Checks whether the RadioGroup control or a specific element is visible."""
        return self.radiogroup.isVisible(id)

    def set_properties(self, arg: Union[str, Dict[str, Any]], props: Dict[str, Any] = None) -> None:
        """Allows changing configuration attributes of the control dynamically."""
        if isinstance(arg, str):
            # arg is id, props is properties
            self.radiogroup.setProperties(arg, js.JSON.parse(json.dumps(props)))
        else:
            # arg is properties, props is None
            self.radiogroup.setProperties(js.JSON.parse(json.dumps(arg)))

    def set_value(self, value: str) -> None:
        """Sets the value for the RadioGroup control."""
        self.radiogroup.setValue(value)

    def show(self, id: str = None) -> None:
        """Shows either a radio button of RadioGroup or the whole RadioGroup."""
        self.radiogroup.show(id)

    def validate(self, silent: bool = False) -> bool:
        """Validates the RadioGroup control."""
        return self.radiogroup.validate(silent)

    """ RadioGroup Events """

    def add_event_handler(self, event_name: str, handler: Callable) -> None:
        """Helper to add event handlers dynamically."""
        event_proxy = create_proxy(handler)
        self.radiogroup.events.on(event_name, event_proxy)

    def on_after_change_properties(self, handler: Callable[[Dict[str, Any]], None]) -> None:
        """Fires after configuration attributes have been changed dynamically."""
        self.add_event_handler('afterChangeProperties', handler)

    def on_after_hide(self, handler: Callable[[str, str, bool], None]) -> None:
        """Fires after the control or its radio button is hidden."""
        self.add_event_handler('afterHide', handler)

    def on_after_show(self, handler: Callable[[str, str], None]) -> None:
        """Fires after the control or its radio button is shown."""
        self.add_event_handler('afterShow', handler)

    def on_after_validate(self, handler: Callable[[str, bool], None]) -> None:
        """Fires after the control value is validated."""
        self.add_event_handler('afterValidate', handler)

    def on_before_change(self, handler: Callable[[str], Union[bool, None]]) -> None:
        """Fires before changing the value of the control."""
        def event_handler(value):
            result = handler(value)
            if result is False:
                return js.Boolean(False)
        self.radiogroup.events.on('beforeChange', create_proxy(event_handler))

    def on_before_change_properties(self, handler: Callable[[Dict[str, Any]], Union[bool, None]]) -> None:
        """Fires before configuration attributes are changed dynamically."""
        def event_handler(properties):
            result = handler(properties.to_py())
            if result is False:
                return js.Boolean(False)
        self.radiogroup.events.on('beforeChangeProperties', create_proxy(event_handler))

    def on_before_hide(self, handler: Callable[[str, str, bool], Union[bool, None]]) -> None:
        """Fires before the control or its radio button is hidden."""
        def event_handler(value, id, init):
            result = handler(value, id, init)
            if result is False:
                return js.Boolean(False)
        self.radiogroup.events.on('beforeHide', create_proxy(event_handler))

    def on_before_show(self, handler: Callable[[str, str], Union[bool, None]]) -> None:
        """Fires before the control or its radio button is shown."""
        def event_handler(value, id):
            result = handler(value, id)
            if result is False:
                return js.Boolean(False)
        self.radiogroup.events.on('beforeShow', create_proxy(event_handler))

    def on_before_validate(self, handler: Callable[[str], Union[bool, None]]) -> None:
        """Fires before the control value is validated."""
        def event_handler(value):
            result = handler(value)
            if result is False:
                return js.Boolean(False)
        self.radiogroup.events.on('beforeValidate', create_proxy(event_handler))

    def on_blur(self, handler: Callable[[str, str], None]) -> None:
        """Fires when the RadioGroup control has lost focus."""
        self.add_event_handler('blur', handler)

    def on_change(self, handler: Callable[[str], None]) -> None:
        """Fires on changing the value of the control."""
        self.add_event_handler('change', handler)

    def on_focus(self, handler: Callable[[str, str], None]) -> None:
        """Fires when the RadioGroup control has received focus."""
        self.add_event_handler('focus', handler)

    def on_keydown(self, handler: Callable[[Any, str], None]) -> None:
        """Fires when any key is pressed and a radio button is in focus."""
        self.add_event_handler('keydown', handler)



from typing import Any, Callable, Dict, Optional, Union



class SliderConfig:
    """
    Configuration class for the Slider widget.
    """

    def __init__(self,
                 min: float,
                 max: float,
                 step: float = 1,
                 value: Union[float, str, list] = None,
                 css: str = None,
                 helpMessage: str = None,
                 hiddenLabel: bool = None,
                 inverse: bool = None,
                 label: str = None,
                 labelPosition: str = None,
                 labelWidth: Union[str, int] = None,
                 majorTick: float = None,
                 mode: str = "horizontal",
                 range: bool = None,
                 tick: float = None,
                 tickTemplate: Callable[[float], str] = None,
                 tooltip: bool = True):
        """
        Initializes the SliderConfig.

        :param min: (Required) The minimal value of slider.
        :param max: (Required) The maximal value of slider.
        :param step: (Optional) The step the slider thumb will be moved with. Default is 1.
        :param value: (Optional) The value the thumb will be set at on initialization of the slider.
        :param css: (Optional) Adds style classes for the component.
        :param helpMessage: (Optional) Adds an icon with a question mark next to Slider.
        :param hiddenLabel: (Optional) Adds a hidden label for a Slider that will be used while sending a form to the server.
        :param inverse: (Optional) Enables/disables the inverse slider mode.
        :param label: (Optional) Specifies the label of a slider.
        :param labelPosition: (Optional) Defines the position of a label of a slider ("left" | "top").
        :param labelWidth: (Optional) Sets the width of a label.
        :param majorTick: (Optional) Sets interval of rendering numeric values on the slider scale.
        :param mode: (Optional) The direction of the Slider scale ("vertical" | "horizontal"). Default is "horizontal".
        :param range: (Optional) Enables/disables the possibility to select a range of values on the slider.
        :param tick: (Optional) Sets the interval of steps for rendering the slider scale.
        :param tickTemplate: (Optional) Sets a template for rendering values on the scale.
        :param tooltip: (Optional) Enables a tooltip on hovering over the slider thumb. Default is True.
        """
        self.min = min
        self.max = max
        self.step = step
        self.value = value
        self.css = css
        self.helpMessage = helpMessage
        self.hiddenLabel = hiddenLabel
        self.inverse = inverse
        self.label = label
        self.labelPosition = labelPosition
        self.labelWidth = labelWidth
        self.majorTick = majorTick
        self.mode = mode
        self.range = range
        self.tick = tick
        self.tickTemplate = tickTemplate
        self.tooltip = tooltip

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the SliderConfig into a dictionary format.
        """
        config_dict = {
            'min': self.min,
            'max': self.max,
            'step': self.step,
            'value': self.value,
            'css': self.css,
            'helpMessage': self.helpMessage,
            'hiddenLabel': self.hiddenLabel,
            'inverse': self.inverse,
            'label': self.label,
            'labelPosition': self.labelPosition,
            'labelWidth': self.labelWidth,
            'majorTick': self.majorTick,
            'mode': self.mode,
            'range': self.range,
            'tick': self.tick,
            'tooltip': self.tooltip,
        }
        # Handle the tickTemplate function
        if self.tickTemplate:
            # Since tickTemplate is a function, we can't serialize it to JSON
            # We'll store it separately and handle it in the Slider class
            config_dict['tickTemplate'] = self.tickTemplate

        # Remove None values
        config_dict = {k: v for k, v in config_dict.items() if v is not None}
        return config_dict



from .sidebar import Sidebar
from .sidebar_config import SidebarConfig, CustomHTMLConfig, MenuItemConfig, NavItemConfig, SeparatorConfig, SpacerConfig, TitleConfig

__all__ = [
    'Sidebar',
    'SidebarConfig',
    'CustomHTMLConfig',
    'MenuItemConfig',
    'NavItemConfig',
    'SeparatorConfig',
    'SpacerConfig',
    'TitleConfig'
]


from typing import List, Union, Dict, Any, Callable

class ControlConfig:
    def to_dict(self) -> Dict[str, Any]:
        return {k: v for k, v in self.__dict__.items() if v is not None}

class CalendarConfig(ControlConfig):
    """Configuration class for Calendar."""
    def __init__(self,
                 date: Union[str, Any] = None,
                 date_format: str = "%d/%m/%y",
                 disabled_dates: Callable[[Any], bool] = None,
                 mark: Callable[[Any], str] = None,
                 mode: str = "calendar",
                 range: bool = False,
                 this_month_only: bool = False,
                 time_format: int = 24,
                 time_picker: bool = False,
                 value: Union[str, Any, List[Union[str, Any]]] = None,
                 week_numbers: bool = False,
                 week_start: str = "sunday",
                 css: str = None,
                 width: Union[int, str] = "250px"):
        self.date = date
        self.date_format = date_format
        self.disabled_dates = disabled_dates
        self.mark = mark
        self.mode = mode
        self.range = range
        self.this_month_only = this_month_only
        self.time_format = time_format
        self.time_picker = time_picker
        self.value = value
        self.week_numbers = week_numbers
        self.week_start = week_start
        self.css = css
        self.width = width


from typing import Any, Dict, Optional



class PaginationConfig:
    """
    Configuration class for the Pagination widget.
    """
    def __init__(self,
                 data: Any,  # Should be a DataCollection instance from DHTMLX library
                 css: str = None,
                 inputWidth: int = 40,
                 page: int = 0,
                 pageSize: int = 10):
        """
        Initializes the PaginationConfig.

        :param data: (Required) The data collection of a widget to set into the pagination.
        :param css: (Optional) Adds style classes to the pagination.
        :param inputWidth: (Optional) Sets the width for the input of the pagination.
        :param page: (Optional) The index of the initial page set in the pagination.
        :param pageSize: (Optional) The number of items displayed per page of the related widget.
        """
        self.data = data
        self.css = css
        self.inputWidth = inputWidth
        self.page = page
        self.pageSize = pageSize

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the PaginationConfig into a dictionary format.
        """
        config_dict = {
            'data': self.data,
            'css': self.css,
            'inputWidth': self.inputWidth,
            'page': self.page,
            'pageSize': self.pageSize,
        }
        # Remove None values
        return {k: v for k, v in config_dict.items() if v is not None}



"""
pyTincture Tabbar widget implementation
"""

from typing import Any, Callable, Dict, List, Union
import json
from pyodide.ffi import create_proxy
import js
from uuid import uuid4

from .tabbar_config import TabbarConfig

from ..grid import Grid, GridConfig
from ..toolbar import Toolbar, ToolbarConfig
from ..sidebar import Sidebar, SidebarConfig
from ..form import Form, FormConfig
from ..menu import Menu, MenuConfig
#from ..layout import Layout, LayoutConfig, CellConfig
from ..listbox import Listbox, ListboxConfig
from ..calendar import Calendar, CalendarConfig
from ..chart import Chart, ChartConfig
from ..pagination import Pagination, PaginationConfig
from ..ribbon import Ribbon, RibbonConfig
from .tabbar_config import TabbarConfig, TabConfig
from ..timepicker import Timepicker, TimepickerConfig
from ..tree import Tree, TreeConfig
from ..kanban import Kanban, KanbanConfig
from ..cardflow import CardFlow, CardFlowConfig
from ..cardpanel import CardPanel, CardPanelConfig



class Tabbar:
    def __init__(self, config: TabbarConfig = None, widget_parent: Any = None):
        """Initializes the Tabbar instance."""
        if config is None:
            config = TabbarConfig()
        config_dict = config.to_dict()
        self.tabbar = js.dhx.Tabbar.new(widget_parent, js.JSON.parse(json.dumps(config_dict)))

    """ Placeholder Widgets Adders """

    def add_grid(self, id: str = "mainwindow", grid_config: GridConfig = None) -> Grid:
        """Adds a Grid widget into a Layout cell."""
        grid_widget = Grid(config=grid_config)
        self.attach(id, grid_widget.grid)
        if grid_config.data:
            grid_widget.grid.data.parse(js.JSON.parse(json.dumps(grid_config.data)))
        return grid_widget
    
    def add_cardflow(self, id: str, cardflow_config: CardFlowConfig = None) -> Any:
        """Adds a CardFlow widget into a Layout cell."""
        cardflow_widget = CardFlow(config=cardflow_config, container=self.tabbar.getCell(id))
        return cardflow_widget
    
    def add_cardpanel(self, id: str, cardpanel_config: CardPanelConfig = None) -> CardPanel:
        """Adds a CardPanel widget into a Tabbar cell."""
        cardpanel_widget = CardPanel(config=cardpanel_config, container=self.tabbar.getCell(id))
        return cardpanel_widget
    
    #def add_layout(self, id: str = "mainwindow", layout_config: LayoutConfig = None) -> TLayout:
    #    """ Adds a Layout into a Layout cell """
    #    layout_widget = Layout(config=layout_config)
    #    self.attach(id, layout_widget.layout)
    #    return layout_widget

        # Define a function to wait for the element to be ready
    def wait_for_element(self, selector, callback):
        def check_element():
            if js.document.querySelector(selector):
                callback()
            else:
                js.window.setTimeout(create_proxy(check_element), 100)  # Check again after 100ms

        check_element()

    def add_kanban(self, id: str = "mainwindow", kanban_config: KanbanConfig = None, kanban_callback: callable = None) -> None:
        self.kanban_callback = kanban_callback
        self.kanban_config = kanban_config
        self.kanban_div_id = "kanban_root_" + str(uuid4()).replace("-", "") #"kanban_root"
        self.attach_html(id, f'<div id="{self.kanban_div_id}" style="width:100%;height:100%;"></div>')
        self.wait_for_element(f"#{self.kanban_div_id}", self.create_kanban)

    def create_kanban(self):
        return_kanban = Kanban(self.kanban_config, f"#{self.kanban_div_id}")
        current_theme = js.document.documentElement.getAttribute('data-dhx-theme')
        if current_theme == "dark":
            theme = "willow-dark"
        else:
            theme = "willow"

        return_kanban.kanban.setTheme(js.JSON.parse(json.dumps({"name": theme, "fonts": True})))
        self.kanban_callback(return_kanban)
    
    def add_menu(self, id: str = "mainwindow_header", menu_config: MenuConfig = None) -> Menu:
        """ Adds a Layout into a Layout cell """
        menu_widget = Menu(config=menu_config)
        self.attach(id, menu_widget.menu)
        return menu_widget

    def add_toolbar(self, id: str = "mainwindow", toolbar_config: ToolbarConfig = None) -> Toolbar:
        """Adds a Toolbar widget into a Layout cell."""
        toolbar_widget = Toolbar(config=toolbar_config)
        self.attach(id, toolbar_widget.toolbar)
        return toolbar_widget

    def add_sidebar(self, id: str, sidebar_config: SidebarConfig = None) -> Sidebar:
        """Adds a Sidebar widget into a Layout cell."""
        sidebar_widget = Sidebar(config=sidebar_config)
        self.attach(id, sidebar_widget.sidebar)
        return sidebar_widget

    def add_form(self, id: str, form_config: FormConfig = None) -> Form:
        """Adds a Form widget into a Layout cell."""
        form_widget = Form(config=form_config)
        self.attach(id, form_widget.form)
        return form_widget
    
    def add_listbox(self, id: str, listbox_config: ListboxConfig = None) -> Any:
        """Adds a Listbox widget into a Layout cell."""
        listbox_widget = Listbox(config=listbox_config)
        self.attach(id, listbox_widget.listbox)
        return listbox_widget
    
    def add_calendar(self, id: str, calendar_config: CalendarConfig = None) -> Any:
        """Adds a Calendar widget into a Layout cell."""
        calendar_widget = Calendar(config=calendar_config)
        self.attach(id, calendar_widget.calendar)
        return calendar_widget
    
    def add_chart(self, id: str, chart_config: ChartConfig = None) -> Any:
        """Adds a Chart widget into a Layout cell."""
        chart_widget = Chart(config=chart_config)
        self.attach(id, chart_widget.chart)
        return chart_widget
    
    def add_pagination(self, id: str, pagination_config: PaginationConfig = None) -> Any:
        """Adds a Pagination widget into a Layout cell."""
        pagination_widget = Pagination(config=pagination_config)
        self.attach(id, pagination_widget.pagination)
        return pagination_widget
    
    def add_ribbon(self, id: str, ribbon_config: RibbonConfig = None) -> Any:
        """Adds a Ribbon widget into a Layout cell."""
        ribbon_widget = Ribbon(config=ribbon_config)
        self.attach(id, ribbon_widget.ribbon)
        return ribbon_widget
    
    def add_tabbar(self, id: str, tabbar_config: TabbarConfig = None) -> Any:
        """Adds a Tabbar widget into a Layout cell."""
        tabbar_widget = Tabbar(config=tabbar_config)
        self.attach(id, tabbar_widget.tabbar)
        return tabbar_widget
    
    def add_timepicker(self, id: str, timepicker_config: TimepickerConfig = None) -> Any:
        """Adds a Timepicker widget into a Layout cell."""
        timepicker_widget = Timepicker(config=timepicker_config)
        self.attach(id, timepicker_widget.timepicker)
        return timepicker_widget
    
    def add_tree(self, id: str, tree_config: TreeConfig = None) -> Any:
        """Adds a Tree widget into a Layout cell."""
        tree_widget = Tree(config=tree_config)
        self.attach(id, tree_widget.tree)
        return tree_widget

    """ Tabbar API Methods """

    def add_tab(self, config: Dict[str, Any], index: int) -> None:
        """Adds a new tab into a tabbar."""
        self.tabbar.addTab(js.JSON.parse(json.dumps(config)), index)

    def destructor(self) -> None:
        """Removes the Tabbar instance and releases occupied resources."""
        self.tabbar.destructor()

    def disable_tab(self, id: str) -> bool:
        """Disables a tab on a page."""
        return self.tabbar.disableTab(id)

    def enable_tab(self, id: str) -> None:
        """Enables a disabled tab."""
        self.tabbar.enableTab(id)

    def get_active(self) -> str:
        """Gets the id of the active tab."""
        return self.tabbar.getActive()

    def get_cell(self, id: str) -> Any:
        """Returns the config object of a cell."""
        return self.tabbar.getCell(id)

    def get_id(self, index: int) -> str:
        """Returns the id of a tab by its index."""
        return self.tabbar.getId(index)

    def get_widget(self) -> Any:
        """Returns the widget attached to Tabbar."""
        return self.tabbar.getWidget()

    def is_disabled(self, id: str) -> bool:
        """Checks whether a tab is disabled."""
        return self.tabbar.isDisabled(id)

    def paint(self) -> None:
        """Repaints the Tabbar on a page."""
        self.tabbar.paint()

    def remove_tab(self, id: str) -> None:
        """Removes a tab from a tabbar."""
        self.tabbar.removeTab(id)

    def set_active(self, id: str) -> None:
        """Sets an active tab."""
        self.tabbar.setActive(id)

    """ Tabbar Events """

    def add_event_handler(self, event_name: str, handler: Callable) -> None:
        """Helper to add event handlers dynamically."""
        event_proxy = create_proxy(handler)
        self.tabbar.events.on(event_name, event_proxy)

    def on_after_close(self, handler: Callable[[str], None]) -> None:
        """Fires after closing a tab in Tabbar."""
        self.add_event_handler('afterClose', handler)

    def on_before_change(self, handler: Callable[[str, str], Union[bool, None]]) -> None:
        """Fires before changing the active tab."""
        def event_handler(id, prev):
            result = handler(id, prev)
            if result is False:
                return js.Boolean(False)
        event_proxy = create_proxy(event_handler)
        self.tabbar.events.on('beforeChange', event_proxy)

    def on_before_close(self, handler: Callable[[str], Union[bool, None]]) -> None:
        """Fires before closing a tab in Tabbar."""
        def event_handler(id):
            result = handler(id)
            if result is False:
                return js.Boolean(False)
        event_proxy = create_proxy(event_handler)
        self.tabbar.events.on('beforeClose', event_proxy)

    def on_change(self, handler: Callable[[str, str], None]) -> None:
        """Fires on changing the active tab."""
        self.add_event_handler('change', handler)
        return True

    """ Tabbar Properties """

    @property
    def active_tab(self) -> str:
        """Gets or sets the currently active tab."""
        return self.tabbar.config.activeTab

    @active_tab.setter
    def active_tab(self, value: str) -> None:
        self.tabbar.config.activeTab = value

    @property
    def closable(self) -> Union[bool, List[str]]:
        """Gets or sets the closable property for tabs."""
        return self.tabbar.config.closable

    @closable.setter
    def closable(self, value: Union[bool, List[str]]) -> None:
        self.tabbar.config.closable = value

    @property
    def css(self) -> str:
        """Gets or sets the CSS class(es) applied to Tabbar."""
        return self.tabbar.config.css

    @css.setter
    def css(self, value: str) -> None:
        self.tabbar.config.css = value

    @property
    def disabled(self) -> Union[str, List[str]]:
        """Gets or sets disabled tabs."""
        return self.tabbar.config.disabled

    @disabled.setter
    def disabled(self, value: Union[str, List[str]]) -> None:
        self.tabbar.config.disabled = value

    @property
    def mode(self) -> str:
        """Gets or sets the mode of displaying a tabbar."""
        return self.tabbar.config.mode

    @mode.setter
    def mode(self, value: str) -> None:
        self.tabbar.config.mode = value

    @property
    def no_content(self) -> bool:
        """Gets or sets whether tabs contain any content."""
        return self.tabbar.config.noContent

    @no_content.setter
    def no_content(self, value: bool) -> None:
        self.tabbar.config.noContent = value

    @property
    def tab_align(self) -> str:
        """Gets or sets alignment for tabs."""
        return self.tabbar.config.tabAlign

    @tab_align.setter
    def tab_align(self, value: str) -> None:
        self.tabbar.config.tabAlign = value

    @property
    def tab_auto_height(self) -> bool:
        """Gets or sets whether the height of tabs is automatically adjusted."""
        return self.tabbar.config.tabAutoHeight

    @tab_auto_height.setter
    def tab_auto_height(self, value: bool) -> None:
        self.tabbar.config.tabAutoHeight = value

    @property
    def tab_auto_width(self) -> bool:
        """Gets or sets whether the width of tabs is automatically adjusted."""
        return self.tabbar.config.tabAutoWidth

    @tab_auto_width.setter
    def tab_auto_width(self, value: bool) -> None:
        self.tabbar.config.tabAutoWidth = value

    @property
    def tab_height(self) -> Union[int, str]:
        """Gets or sets the height of a tab."""
        return self.tabbar.config.tabHeight

    @tab_height.setter
    def tab_height(self, value: Union[int, str]) -> None:
        self.tabbar.config.tabHeight = value

    @property
    def tab_width(self) -> Union[int, str]:
        """Gets or sets the width of a tab."""
        return self.tabbar.config.tabWidth

    @tab_width.setter
    def tab_width(self, value: Union[int, str]) -> None:
        self.tabbar.config.tabWidth = value

    @property
    def views(self) -> List[Dict[str, Any]]:
        """Gets or sets the configuration of tabs."""
        return self.tabbar.config.views

    @views.setter
    def views(self, value: List[Dict[str, Any]]) -> None:
        self.tabbar.config.views = value

    """ Tabbar Cell API Functions """

    def attach(self, id: str, component: Union[str, Any], config: Dict[str, Any] = None) -> Any:
        """Attaches a DHTMLX component into a Tabbar cell."""
        return self.tabbar.getCell(id).attach(component, js.JSON.parse(json.dumps(config or {})))

    def attach_html(self, id: str, html: str) -> None:
        """Adds HTML content into a Tabbar cell."""
        self.tabbar.getCell(id).attachHTML(html)

    def get_cell_parent(self, id: str) -> Any:
        """Returns the parent of a cell."""
        return self.tabbar.getCell(id).getParent()

    def get_cell_widget(self, id: str) -> Any:
        """Returns the widget attached to a Tabbar cell."""
        return self.tabbar.getCell(id).getWidget()



from typing import Any, Dict, Optional, Union


class WindowConfig:
    """
    Configuration class for the Window widget.
    """
    def __init__(self,
                 closable: Optional[bool] = False,
                 css: Optional[str] = None,
                 footer: Optional[bool] = None,
                 header: Optional[bool] = None,
                 height: Optional[Union[int, str]] = "50%",
                 html: Optional[str] = None,
                 left: Optional[int] = None,
                 minHeight: Optional[int] = 100,
                 minWidth: Optional[int] = 100,
                 modal: Optional[bool] = False,
                 movable: Optional[bool] = False,
                 node: Optional[Union[Any, str]] = None,
                 resizable: Optional[bool] = False,
                 title: Optional[str] = None,
                 top: Optional[int] = None,
                 viewportOverflow: Optional[bool] = False,
                 width: Optional[Union[int, str]] = "50%"):
        """
        Initializes the WindowConfig.

        :param closable: Defines whether a window can be closed.
        :param css: Adds style classes for the window.
        :param footer: Adds a footer to the window.
        :param header: Adds a header to the window.
        :param height: Sets the height of the window.
        :param html: Sets an HTML content into a window on initialization.
        :param left: The left coordinate of a window position.
        :param minHeight: Sets the minimal height of a window.
        :param minWidth: Sets the minimal width of a window.
        :param modal: Defines whether a window is modal.
        :param movable: Defines whether a window is movable.
        :param node: The container for a window or its ID.
        :param resizable: Defines whether a window can be resized.
        :param title: Adds some text into the header of a window.
        :param top: The top coordinate of a window position.
        :param viewportOverflow: Defines whether a window can go beyond borders of the browser.
        :param width: Sets the width of the window.
        """
        self.closable = closable
        self.css = css
        self.footer = footer
        self.header = header
        self.height = height
        self.html = html
        self.left = left
        self.minHeight = minHeight
        self.minWidth = minWidth
        self.modal = modal
        self.movable = movable
        self.node = node
        self.resizable = resizable
        self.title = title
        self.top = top
        self.viewportOverflow = viewportOverflow
        self.width = width

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the WindowConfig into a dictionary format.
        """
        config_dict = {
            'closable': self.closable,
            'css': self.css,
            'footer': self.footer,
            'header': self.header,
            'height': self.height,
            'html': self.html,
            'left': self.left,
            'minHeight': self.minHeight,
            'minWidth': self.minWidth,
            'modal': self.modal,
            'movable': self.movable,
            'node': self.node,
            'resizable': self.resizable,
            'title': self.title,
            'top': self.top,
            'viewportOverflow': self.viewportOverflow,
            'width': self.width
        }
        return {k: v for k, v in config_dict.items() if v is not None}



from typing import Any, Dict, List, Optional, Union


class ControlConfig:
    def to_dict(self) -> Dict[str, Any]:
        return {k: v for k, v in self.__dict__.items() if v is not None}


class CustomHTMLConfig(ControlConfig):
    def __init__(self,
                 id: Optional[str] = None,
                 parent: Optional[str] = None,
                 html: Optional[str] = None,
                 css: Optional[str] = None,
                 hidden: Optional[bool] = None):
        self.type = "customHTML"
        self.id = id
        self.parent = parent
        self.html = html
        self.css = css
        self.hidden = hidden


class MenuItemConfig(ControlConfig):
    def __init__(self,
                 id: Optional[str] = None,
                 parent: Optional[str] = None,
                 value: Optional[str] = None,
                 items: Optional[List['ControlConfig']] = None,
                 count: Optional[int] = None,
                 countColor: Optional[str] = None,
                 hotkey: Optional[str] = None,
                 html: Optional[str] = None,
                 icon: Optional[str] = None,
                 tooltip: Optional[str] = None,
                 css: Optional[str] = None,
                 disabled: Optional[bool] = None,
                 hidden: Optional[bool] = None):
        self.type = "menuItem"
        self.id = id
        self.parent = parent
        self.value = value
        self.items = items
        self.count = count
        self.countColor = countColor
        self.hotkey = hotkey
        self.html = html
        self.icon = icon
        self.tooltip = tooltip
        self.css = css
        self.disabled = disabled
        self.hidden = hidden

    def to_dict(self) -> Dict[str, Any]:
        config = {k: v for k, v in self.__dict__.items() if v is not None}
        if self.items:
            config['items'] = [item.to_dict() for item in self.items]
        return config


class NavItemConfig(ControlConfig):
    def __init__(self,
                 id: Optional[str] = None,
                 parent: Optional[str] = None,
                 value: Optional[str] = None,
                 items: Optional[List['ControlConfig']] = None,
                 active: Optional[bool] = None,
                 count: Optional[int] = None,
                 countColor: Optional[str] = None,
                 group: Optional[str] = None,
                 hotkey: Optional[str] = None,
                 html: Optional[str] = None,
                 icon: Optional[str] = None,
                 tooltip: Optional[str] = None,
                 twoState: Optional[bool] = None,
                 css: Optional[str] = None,
                 disabled: Optional[bool] = None,
                 hidden: Optional[bool] = None):
        self.type = "navItem"
        self.id = id
        self.parent = parent
        self.value = value
        self.items = items
        self.active = active
        self.count = count
        self.countColor = countColor
        self.group = group
        self.hotkey = hotkey
        self.html = html
        self.icon = icon
        self.tooltip = tooltip
        self.twoState = twoState
        self.css = css
        self.disabled = disabled
        self.hidden = hidden

    def to_dict(self) -> Dict[str, Any]:
        config = {k: v for k, v in self.__dict__.items() if v is not None}
        if self.items:
            config['items'] = [item.to_dict() for item in self.items]
        return config


class SeparatorConfig(ControlConfig):
    def __init__(self,
                 id: Optional[str] = None):
        self.type = "separator"
        self.id = id


class SpacerConfig(ControlConfig):
    def __init__(self,
                 id: Optional[str] = None):
        self.type = "spacer"
        self.id = id


class TitleConfig(ControlConfig):
    def __init__(self,
                 id: Optional[str] = None,
                 parent: Optional[str] = None,
                 value: Optional[str] = None,
                 html: Optional[str] = None,
                 tooltip: Optional[str] = None,
                 css: Optional[str] = None,
                 disabled: Optional[bool] = None,
                 hidden: Optional[bool] = None):
        self.type = "title"
        self.id = id
        self.parent = parent
        self.value = value
        self.html = html
        self.tooltip = tooltip
        self.css = css
        self.disabled = disabled
        self.hidden = hidden


class SidebarConfig:
    """
    Configuration class for the Sidebar widget.
    """
    def __init__(self,
                 data: Optional[List[ControlConfig]] = None,
                 collapsed: Optional[bool] = None,
                 css: Optional[str] = None,
                 menuCss: Optional[str] = None,
                 minWidth: Union[int, str] = None,
                 width: Union[int, str] = None):
        """
        Initializes the SidebarConfig.

        :param data: (Optional) List of control configurations to set into Sidebar.
        :param collapsed: (Optional) Defines that a sidebar is initialized in the collapsed state.
        :param css: (Optional) Adds style classes to Sidebar.
        :param menuCss: (Optional) Adds style classes to all containers of Sidebar controls with nested items.
        :param minWidth: (Optional) Sets the minimal width of a sidebar in the collapsed state.
        :param width: (Optional) Sets the width of a sidebar.
        """
        self.data = data
        self.collapsed = collapsed
        self.css = css
        self.menuCss = menuCss
        self.minWidth = minWidth
        self.width = width

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the SidebarConfig into a dictionary format.
        """
        data_dicts = [item.to_dict() for item in self.data] if self.data else None
        config_dict = {
            'data': data_dicts,
            'collapsed': self.collapsed,
            'css': self.css,
            'menuCss': self.menuCss,
            'minWidth': self.minWidth,
            'width': self.width
        }
        return {k: v for k, v in config_dict.items() if v is not None}



"""
pyTincture layout widget implementation
"""

from typing import Any, Callable, Dict, List, Union, TypeVar
import json
from pyodide.ffi import create_proxy
import js
from uuid import uuid4

from ..grid import Grid, GridConfig
from ..toolbar import Toolbar, ToolbarConfig
from ..sidebar import Sidebar, SidebarConfig
from ..form import Form, FormConfig
from ..menu import Menu, MenuConfig
from .layout_config import LayoutConfig, CellConfig
from ..listbox import Listbox, ListboxConfig
from ..calendar import Calendar, CalendarConfig
from ..chart import Chart, ChartConfig
from ..pagination import Pagination, PaginationConfig
from ..ribbon import Ribbon, RibbonConfig
from ..tabbar import Tabbar, TabbarConfig
from ..timepicker import Timepicker, TimepickerConfig
from ..tree import Tree, TreeConfig
from ..kanban import Kanban, KanbanConfig
from ..cardflow import CardFlow, CardFlowConfig
from ..cardpanel import CardPanel, CardPanelConfig
from ..chat import Chat, ChatConfig


TLayout = TypeVar("TLayout", bound="Layout")

class LoadUICaller(type):
    def __call__(cls, *args, **kwargs):
        """Called when you call MyNewClass() """
        obj = type.__call__(cls, *args, **kwargs)
        obj.load_ui()
        return obj

class Layout(object, metaclass=LoadUICaller):
    layout_config = None
    def __init__(self, config: LayoutConfig = None, mainwindow=False, **kwargs):
        self.parent = kwargs.get("parent", None)
        """Initializes the layout instance."""
        mainconfig = self.layout_config or config or {}
        mainconfig = mainconfig if type(mainconfig) is dict else mainconfig.to_dict()
        if not mainconfig:
            mainconfig = LayoutConfig(
                css="dhx_layout-cell--bordered",
                type="line",
                rows=[
                    CellConfig(id="mainwindow_header", width="98vw", height="auto", header=None),
                    CellConfig(id="mainwindow", width="98vw", header=None)
                ]
            ).to_dict()

        if mainwindow:
            self.layout = js.dhx.Layout.new("maindiv", js.JSON.parse(json.dumps(mainconfig)))
        else:
            self.layout = js.dhx.Layout.new(None, js.JSON.parse(json.dumps(mainconfig)))
            
        self.initialized = False

    """ Placeholder Widgets Adders """

    def load_ui(self, *args, **kwargs):
        """Subclass this to build your UI"""
        pass

    def add_grid(self, id: str = "mainwindow", grid_config: GridConfig = None) -> Grid:
        """Adds a Grid widget into a Layout cell."""
        grid_widget = Grid(config=grid_config)
        self.attach(id, grid_widget.grid)
        if grid_config.data:
            grid_widget.grid.data.removeAll()
            grid_widget.grid.data.parse(js.JSON.parse(json.dumps(grid_config.data)))
        return grid_widget

    # Define a function to wait for the element to be ready
    def wait_for_element(self, selector, callback):
        def check_element():
            if js.document.querySelector(selector):
                callback()
            else:
                js.window.setTimeout(create_proxy(check_element), 100)  # Check again after 100ms

        check_element()

    def add_kanban(self, id: str = "mainwindow", kanban_config: KanbanConfig = None, kanban_callback: callable = None) -> None:
        self.kanban_callback = kanban_callback
        self.kanban_config = kanban_config
        self.kanban_div_id = "kanban_root_" + str(uuid4()).replace("-", "")
        self.attach_html(id, f'<div id="{self.kanban_div_id}" style="width:100%;height:100%;"></div>')
        self.wait_for_element(f"#{self.kanban_div_id}", self.create_kanban)

    def create_kanban(self):
        return_kanban = Kanban(config=self.kanban_config, root=f"#{self.kanban_div_id}")
        current_theme = js.document.documentElement.getAttribute('data-dhx-theme')
        if current_theme == "dark":
            theme = "dark"
        else:
            theme = "light"

        return_kanban.set_theme(theme)
        if self.kanban_callback:
            self.kanban_callback(return_kanban)
    
    def add_layout(self, id: str = "mainwindow", layout_config: LayoutConfig = None) -> TLayout:
        """ Adds a Layout into a Layout cell """
        layout_widget = Layout(config=layout_config)
        self.attach(id, layout_widget.layout)
        return layout_widget
    
    def add_menu(self, id: str = "mainwindow_header", menu_config: MenuConfig = None) -> Menu:
        """ Adds a Layout into a Layout cell """
        menu_widget = Menu(config=menu_config)
        self.attach(id, menu_widget.menu)
        return menu_widget

    def add_toolbar(self, id: str = "mainwindow", toolbar_config: ToolbarConfig = None) -> Toolbar:
        """Adds a Toolbar widget into a Layout cell."""
        toolbar_widget = Toolbar(config=toolbar_config)
        self.attach(id, toolbar_widget.toolbar)
        return toolbar_widget

    def add_sidebar(self, id: str, sidebar_config: SidebarConfig = None) -> Sidebar:
        """Adds a Sidebar widget into a Layout cell."""
        sidebar_widget = Sidebar(config=sidebar_config)
        self.attach(id, sidebar_widget.sidebar)
        return sidebar_widget

    def add_form(self, id: str, form_config: FormConfig = None) -> Form:
        """Adds a Form widget into a Layout cell."""
        form_widget = Form(config=form_config)
        self.attach(id, form_widget.form)
        return form_widget
    
    def add_listbox(self, id: str, listbox_config: ListboxConfig = None) -> Any:
        """Adds a Listbox widget into a Layout cell."""
        listbox_widget = Listbox(config=listbox_config)
        self.attach(id, listbox_widget.listbox)
        return listbox_widget
    
    def add_calendar(self, id: str, calendar_config: CalendarConfig = None) -> Any:
        """Adds a Calendar widget into a Layout cell."""
        calendar_widget = Calendar(config=calendar_config)
        self.attach(id, calendar_widget.calendar)
        return calendar_widget
    
    def add_chart(self, id: str, chart_config: ChartConfig = None) -> Any:
        """Adds a Chart widget into a Layout cell."""
        chart_widget = Chart(config=chart_config)
        self.attach(id, chart_widget.chart)
        return chart_widget
    
    def add_pagination(self, id: str, pagination_config: PaginationConfig = None) -> Any:
        """Adds a Pagination widget into a Layout cell."""
        pagination_widget = Pagination(config=pagination_config)
        self.attach(id, pagination_widget.pagination)
        return pagination_widget
    
    def add_cardflow(self, id: str, cardflow_config: CardFlowConfig = None) -> Any:
        """Adds a CardFlow widget into a Layout cell."""
        cardflow_widget = CardFlow(config=cardflow_config, container=self.layout.getCell(id))
        return cardflow_widget
    
    def add_cardpanel(self, id: str, cardpanel_config: CardPanelConfig = None) -> CardPanel:
        """Adds a CardPanel widget into a Layout cell."""
        cardpanel_widget = CardPanel(config=cardpanel_config, container=self.layout.getCell(id))
        return cardpanel_widget
    
    def add_chat(self, id: str, chat_config: ChatConfig = None) -> Chat:
        """Adds a Chat widget into a Layout cell."""
        chat_widget = Chat(config=chat_config, container=self.layout.getCell(id))
        return chat_widget
    
    def add_ribbon(self, id: str, ribbon_config: RibbonConfig = None) -> Any:
        """Adds a Ribbon widget into a Layout cell."""
        ribbon_widget = Ribbon(config=ribbon_config)
        self.attach(id, ribbon_widget.ribbon)
        return ribbon_widget
    
    def add_tabbar(self, id: str, tabbar_config: TabbarConfig = None) -> Any:
        """Adds a Tabbar widget into a Layout cell."""
        tabbar_widget = Tabbar(config=tabbar_config)
        self.attach(id, tabbar_widget.tabbar)
        return tabbar_widget
    
    def add_timepicker(self, id: str, timepicker_config: TimepickerConfig = None) -> Any:
        """Adds a Timepicker widget into a Layout cell."""
        timepicker_widget = Timepicker(config=timepicker_config)
        self.attach(id, timepicker_widget.timepicker)
        return timepicker_widget
    
    def add_tree(self, id: str, tree_config: TreeConfig = None) -> Any:
        """Adds a Tree widget into a Layout cell."""
        tree_widget = Tree(config=tree_config)
        self.attach(id, tree_widget.tree)
        return tree_widget
        
    """ Layout API Functions """

    def destructor(self) -> None:
        """Destroys the layout instance."""
        self.layout.destructor()

    def for_each(self, callback: Callable[[Any, int, List[Any]], Any], parent_id: str = None, level: int = None) -> None:
        """Iterates over layout cells, applying the callback function to each."""
        proxy_callback = create_proxy(callback)
        self.layout.forEach(proxy_callback, parent_id, level)

    def get_cell(self, id: str) -> Any:
        """Retrieves a specific layout cell by its ID."""
        return self.layout.getCell(id)

    def progress_hide(self) -> None:
        """Hides the loading progress bar in the layout."""
        self.layout.progressHide()

    def progress_show(self) -> None:
        """Displays the loading progress bar in the layout."""
        self.layout.progressShow()

    def remove_cell(self, id: str) -> None:
        """Removes a specific layout cell by ID."""
        self.layout.removeCell(id)

    def resize(self, id: str) -> None:
        """Manually triggers a resize on a cell."""
        self.layout.resize(id)

    """ Layout API Events """

    def add_event_handler(self, event_name: str, handler: Callable) -> None:
        """Helper to add event handlers dynamically."""
        event_proxy = create_proxy(handler)
        self.layout.events[event_name] = event_proxy

    def after_add(self, handler: Callable) -> None:
        """Fires after a cell is added."""
        self.add_event_handler('afterAdd', handler)

    def after_collapse(self, handler: Callable) -> None:
        """Fires after a cell is collapsed."""
        self.add_event_handler('afterCollapse', handler)

    def after_expand(self, handler: Callable) -> None:
        """Fires after a cell is expanded."""
        self.add_event_handler('afterExpand', handler)

    def after_hide(self, handler: Callable) -> None:
        """Fires after a cell is hidden."""
        self.add_event_handler('afterHide', handler)

    def after_remove(self, handler: Callable) -> None:
        """Fires after a cell is removed."""
        self.add_event_handler('afterRemove', handler)

    def after_resize_end(self, handler: Callable) -> None:
        """Fires after a cell resize ends."""
        self.add_event_handler('afterResizeEnd', handler)

    def after_show(self, handler: Callable) -> None:
        """Fires after a cell is shown."""
        self.add_event_handler('afterShow', handler)

    def before_add(self, handler: Callable) -> None:
        """Fires before adding a cell, returns false to prevent."""
        self.add_event_handler('beforeAdd', handler)

    def before_collapse(self, handler: Callable) -> None:
        """Fires before collapsing a cell, returns false to prevent."""
        self.add_event_handler('beforeCollapse', handler)

    def before_expand(self, handler: Callable) -> None:
        """Fires before expanding a cell, returns false to prevent."""
        self.add_event_handler('beforeExpand', handler)

    def before_hide(self, handler: Callable) -> None:
        """Fires before hiding a cell, returns false to prevent."""
        self.add_event_handler('beforeHide', handler)

    def before_remove(self, handler: Callable) -> None:
        """Fires before removing a cell, returns false to prevent."""
        self.add_event_handler('beforeRemove', handler)

    def before_resize_start(self, handler: Callable) -> None:
        """Fires before resizing a cell, returns false to prevent."""
        self.add_event_handler('beforeResizeStart', handler)

    def before_show(self, handler: Callable) -> None:
        """Fires before showing a cell, returns false to prevent."""
        self.add_event_handler('beforeShow', handler)

    """ Layout API Properties """

    @property
    def cols(self) -> List[Dict[Any, Any]]:
        """Gets or sets the columns of the layout."""
        return self.layout.cols

    @cols.setter
    def cols(self, value: List[Dict[Any, Any]]) -> None:
        self.layout.cols = value

    @property
    def css(self) -> str:
        """Gets or sets the CSS class for the layout."""
        return self.layout.css

    @css.setter
    def css(self, value: str) -> None:
        self.layout.css = value

    @property
    def rows(self) -> List[Dict[Any, Any]]:
        """Gets or sets the rows of the layout."""
        return self.layout.rows

    @rows.setter
    def rows(self, value: List[Dict[Any, Any]]) -> None:
        self.layout.rows = value

    @property
    def type(self) -> str:
        """Gets or sets the layout type ('line', 'wide', 'space', 'none')."""
        return self.layout.type

    @type.setter
    def type(self, value: str) -> None:
        self.layout.type = value

    """ Cell API Functions """

    def attach(self, id: str, component: Union[str, Any], config: Dict[str, Any] = None) -> Any:
        """Attaches a component or HTML content to a cell."""
        return self.layout.getCell(id).attach(component, js.JSON.parse(json.dumps(config or {})))

    def attach_html(self, id: str, html: str) -> None:
        """Inserts HTML content into a cell."""
        self.layout.getCell(id).attachHTML(html)

    def collapse(self, id: str) -> None:
        """Collapses the specified cell."""
        self.layout.getCell(id).collapse()

    def detach(self, id: str) -> None:
        """Removes an attached component or content from a cell."""
        self.layout.getCell(id).detach()

    def expand(self, id: str) -> None:
        """Expands the collapsed cell."""
        self.layout.getCell(id).expand()

    def get_parent(self, id: str) -> Any:
        """Returns the parent cell of the current cell."""
        return self.layout.getCell(id).getParent()

    def get_widget(self, id: str) -> Any:
        """Returns the attached widget in the layout cell."""
        return self.layout.getCell(id).getWidget()

    def hide(self, id: str) -> None:
        """Hides the specified cell."""
        self.layout.getCell(id).hide()

    def is_visible(self, id: str) -> bool:
        """Checks if the cell is visible."""
        return self.layout.getCell(id).isVisible()

    def paint(self) -> None:
        """Repaints the layout."""
        self.layout.paint()

    def toggle(self, id: str) -> None:
        """Toggles between collapsing and expanding the cell."""
        self.layout.getCell(id).toggle()

    """ Cell API Properties """

    @property
    def align(self) -> str:
        """Gets or sets the alignment of the content inside a cell."""
        return self.layout.align

    @align.setter
    def align(self, value: str) -> None:
        self.layout.align = value

    @property
    def resizable(self) -> bool:
        """Gets or sets whether the cell can be resized."""
        return self.layout.resizable

    @resizable.setter
    def resizable(self, value: bool) -> None:
        self.layout.resizable = value

    # Add other properties similarly...

class MainWindow(Layout):
    def __init__(self) -> None:
        """Initialize the Main Window layout."""
        super().__init__(mainwindow=True)
        self.initialized = True
        self.cookie_status = None

    def set_theme(self, theme: str, css_vars=None) -> None:
        """
        Sets the layout theme.

        :param theme: Theme name or dict payload accepted by dhx.setTheme
        :param css_vars: Optional mapping of selectors to CSS custom properties
        """
        try:
            from .. import theme as _theme_helper
            _theme_helper.set_theme(theme, css_vars)
        except Exception:
            js.dhx.setTheme(theme)

    def show_cookie_banner(self):
        # Suppose you've got a <div id="cookie-banner"> in your HTML
        cookie_banner = js.document.getElementById("cookie-banner")
        cookie_banner.style.display = "block"

    def hide_cookie_banner(self):
        cookie_banner = js.document.getElementById("cookie-banner")
        cookie_banner.style.display = "none"

    def accept_cookies(self, _event=None):
        # Set a cookie in the browser to remember consent
        js.document.cookie = "cookie_consent=accepted; path=/; max-age=31536000"
        self.hide_cookie_banner()
        self.cookie_status = True

    def reject_cookies(self, _event=None):
        # Possibly set a cookie or local storage to remember "rejected"
        js.document.cookie = "cookie_consent=rejected; path=/; max-age=31536000"
        self.hide_cookie_banner()
        self.cookie_status = False

    def check_cookie_consent(self):
        # Check the document.cookie in JavaScript from Pyodide
        cookies = js.document.cookie
        if "cookie_consent=accepted" not in cookies and "cookie_consent=rejected" not in cookies:
            self.show_cookie_banner()
            accept_button = js.document.getElementById("accept-btn")
            reject_button = js.document.getElementById("reject-btn")

            accept_button.addEventListener("click", create_proxy(self.accept_cookies))
            reject_button.addEventListener("click", create_proxy(self.reject_cookies))



from typing import Union, Dict, Any



class TextConfig:
    """
    Configuration class for the Text control.
    """
    def __init__(self,
                 name: str = None,
                 id: str = None,
                 value: Union[str, int] = None,
                 css: str = None,
                 disabled: bool = False,
                 height: Union[str, int] = "content",
                 hidden: bool = False,
                 padding: Union[str, int] = "5px",
                 width: Union[str, int] = "content",
                 inputType: str = "text",
                 hiddenLabel: bool = False,
                 label: str = None,
                 labelPosition: str = "top",
                 labelWidth: Union[str, int] = None,
                 helpMessage: str = None,
                 preMessage: str = None,
                 successMessage: str = None,
                 errorMessage: str = None):
        """
        Initializes the TextConfig.

        :param name: (Optional) The name of the control.
        :param id: (Optional) The id of the control.
        :param value: (Optional) The value of the text control.
        :param css: (Optional) Adds style classes to the control.
        :param disabled: (Optional) Whether the control is disabled.
        :param height: (Optional) The height of the control.
        :param hidden: (Optional) Whether the control is hidden.
        :param padding: (Optional) Sets padding between the cell and border.
        :param width: (Optional) The width of the control.
        :param inputType: (Optional) Type of input: "text", "password", "number".
        :param hiddenLabel: (Optional) Makes the label invisible.
        :param label: (Optional) Specifies a label for the control.
        :param labelPosition: (Optional) Position of the label.
        :param labelWidth: (Optional) Width of the label.
        :param helpMessage: (Optional) Adds a help message to the control.
        :param preMessage: (Optional) Instructions for interacting with the control.
        :param successMessage: (Optional) Message after successful validation.
        :param errorMessage: (Optional) Message after validation error.
        """
        self.type = "text"
        self.name = name
        self.id = id
        self.value = value
        self.css = css
        self.disabled = disabled
        self.height = height
        self.hidden = hidden
        self.padding = padding
        self.width = width
        self.inputType = inputType
        self.hiddenLabel = hiddenLabel
        self.label = label
        self.labelPosition = labelPosition
        self.labelWidth = labelWidth
        self.helpMessage = helpMessage
        self.preMessage = preMessage
        self.successMessage = successMessage
        self.errorMessage = errorMessage

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the TextConfig into a dictionary format.
        """
        config_dict = {
            'type': self.type,
            'name': self.name,
            'id': self.id,
            'value': self.value,
            'css': self.css,
            'disabled': self.disabled,
            'height': self.height,
            'hidden': self.hidden,
            'padding': self.padding,
            'width': self.width,
            'inputType': self.inputType,
            'hiddenLabel': self.hiddenLabel,
            'label': self.label,
            'labelPosition': self.labelPosition,
            'labelWidth': self.labelWidth,
            'helpMessage': self.helpMessage,
            'preMessage': self.preMessage,
            'successMessage': self.successMessage,
            'errorMessage': self.errorMessage
        }
        # Remove None values
        return {k: v for k, v in config_dict.items() if v is not None}



from typing import Any, Dict, List, Optional, Union, Callable


class TreeItemConfig:
    def __init__(self,
                 id: Optional[Union[str, int]] = None,
                 value: Optional[str] = None,
                 opened: Optional[bool] = None,
                 checkbox: Optional[bool] = None,
                 items: Optional[List['TreeItemConfig']] = None,
                 icon: Optional[dict] = None):
        """
        Initializes the configuration for a tree item.

        :param id: The ID of the tree item.
        :param value: The value of the tree item.
        :param opened: Whether the tree item is opened by default.
        :param checkbox: Whether the checkbox is displayed for the tree item.
        :param items: Nested child tree items.
        :param icon: Custom icons for the tree item.
        """
        self.id = id
        self.value = value
        self.opened = opened
        self.checkbox = checkbox
        self.items = items
        self.icon = icon

    def to_dict(self) -> Dict[str, Any]:
        config = {
            "id": self.id,
            "value": self.value,
            "opened": self.opened,
            "checkbox": self.checkbox,
            "items": [item.to_dict() for item in self.items] if self.items else None,
            "icon": self.icon
        }
        return {k: v for k, v in config.items() if v is not None}


class TreeConfig:
    """
    Configuration class for the Tree widget.
    """
    def __init__(self,
                 data: Optional[List[TreeItemConfig]] = None,
                 checkbox: Optional[bool] = None,
                 collapsed: Optional[bool] = None,
                 css: Optional[str] = None,
                 dragCopy: Optional[bool] = None,
                 dragMode: Optional[str] = None,
                 dropBehaviour: Optional[str] = None,
                 editable: Optional[bool] = None,
                 icon: Optional[dict] = None,
                 itemHeight: Optional[Union[int, str]] = None,
                 keyNavigation: Optional[bool] = None,
                 rootId: Optional[Union[str, int]] = None,
                 selection: Optional[bool] = None,
                 template: Optional[Callable[[dict, bool], str]] = None):
        """
        Initializes the configuration for the Tree widget.

        :param data: List of TreeItemConfig to define the tree structure.
        :param checkbox: Whether checkboxes are enabled in the tree.
        :param collapsed: Whether the tree is initialized in the collapsed state.
        :param css: Adds custom style classes to the tree.
        :param dragCopy: Defines if the tree item should be copied during drag-and-drop.
        :param dragMode: Defines drag-and-drop mode ("target", "source", "both").
        :param dropBehaviour: Defines the behavior of dragged items ("child", "sibling", "complex").
        :param editable: Enables inline editing of tree items.
        :param icon: Custom icons for tree items.
        :param itemHeight: Sets the height of tree items.
        :param keyNavigation: Enables key navigation for tree.
        :param rootId: The ID for the root tree element.
        :param selection: Enables selection of tree items.
        :param template: Custom template function for rendering tree items.
        """
        self.data = data
        self.checkbox = checkbox
        self.collapsed = collapsed
        self.css = css
        self.dragCopy = dragCopy
        self.dragMode = dragMode
        self.dropBehaviour = dropBehaviour
        self.editable = editable
        self.icon = icon
        self.itemHeight = itemHeight
        self.keyNavigation = keyNavigation
        self.rootId = rootId
        self.selection = selection
        self.template = template

    def to_dict(self) -> Dict[str, Any]:
        config = {
            "data": [item.to_dict() for item in self.data] if self.data else None,
            "checkbox": self.checkbox,
            "collapsed": self.collapsed,
            "css": self.css,
            "dragCopy": self.dragCopy,
            "dragMode": self.dragMode,
            "dropBehaviour": self.dropBehaviour,
            "editable": self.editable,
            "icon": self.icon,
            "itemHeight": self.itemHeight,
            "keyNavigation": self.keyNavigation,
            "rootId": self.rootId,
            "selection": self.selection,
            "template": self.template
        }
        return {k: v for k, v in config.items() if v is not None}



"""
Select control implementation for the Form widget
"""

from typing import Any, Callable, Dict, Union, List
import json
from pyodide.ffi import create_proxy
import js

from .select_config import SelectConfig


class Select:
    def __init__(self, config: SelectConfig = None, widget_parent: Any = None):
        """Initializes the Select control."""
        if config is None:
            config = SelectConfig()
        config_dict = config.to_dict()
        self.select = js.dhx.FormControl.new(widget_parent, js.JSON.parse(json.dumps(config_dict)))

    """ Select API Functions """

    def blur(self) -> None:
        """Removes focus from the Select control."""
        self.select.blur()

    def clear(self) -> None:
        """Clears the value of the Select control."""
        self.select.clear()

    def clear_validate(self) -> None:
        """Clears validation of the Select control."""
        self.select.clearValidate()

    def destructor(self) -> None:
        """Removes the Select instance and releases the occupied resources."""
        self.select.destructor()

    def disable(self, value: Union[str, int] = None) -> None:
        """Disables the Select control or a specific option."""
        self.select.disable(value)

    def enable(self, value: Union[str, int] = None) -> None:
        """Enables the Select control or a specific option."""
        self.select.enable(value)

    def focus(self) -> None:
        """Sets focus to the Select control."""
        self.select.focus()

    def get_properties(self) -> Dict[str, Any]:
        """Returns the configuration attributes of the control."""
        return self.select.getProperties().to_py()

    def get_options(self) -> List[Dict[str, Any]]:
        """Returns an array of Select options."""
        options = self.select.getOptions()
        return [option.to_py() for option in options]

    def get_value(self) -> Union[str, int]:
        """Returns the current value of the Select control."""
        return self.select.getValue()

    def hide(self) -> None:
        """Hides the Select control."""
        self.select.hide()

    def is_disabled(self, value: Union[str, int] = None) -> bool:
        """Checks whether the Select control or a specific option is disabled."""
        return self.select.isDisabled(value)

    def is_visible(self) -> bool:
        """Checks whether the Select control is visible on the page."""
        return self.select.isVisible()

    def set_options(self, options: List[Dict[str, Any]]) -> None:
        """Allows changing a list of Select options dynamically."""
        self.select.setOptions(js.JSON.parse(json.dumps(options)))

    def set_properties(self, properties: Dict[str, Any]) -> None:
        """Allows changing configuration attributes of the control dynamically."""
        self.select.setProperties(js.JSON.parse(json.dumps(properties)))

    def set_value(self, value: Union[str, int]) -> None:
        """Sets the value for the Select control."""
        self.select.setValue(value)

    def show(self) -> None:
        """Shows the Select control on the page."""
        self.select.show()

    def validate(self, silent: bool = False) -> bool:
        """Validates the Select control."""
        return self.select.validate(silent)

    """ Select Events """

    def add_event_handler(self, event_name: str, handler: Callable) -> None:
        """Helper to add event handlers dynamically."""
        event_proxy = create_proxy(handler)
        self.select.events.on(event_name, event_proxy)

    def on_after_change_properties(self, handler: Callable[[Dict[str, Any]], None]) -> None:
        """Fires after configuration attributes have been changed dynamically."""
        self.add_event_handler('afterChangeProperties', handler)

    def on_after_hide(self, handler: Callable[[Union[str, int], bool], None]) -> None:
        """Fires after the control is hidden."""
        self.add_event_handler('afterHide', handler)

    def on_after_show(self, handler: Callable[[Union[str, int]], None]) -> None:
        """Fires after the control is shown."""
        self.add_event_handler('afterShow', handler)

    def on_after_validate(self, handler: Callable[[Union[str, int], bool], None]) -> None:
        """Fires after the control value is validated."""
        self.add_event_handler('afterValidate', handler)

    def on_before_change(self, handler: Callable[[Union[str, int]], Union[bool, None]]) -> None:
        """Fires before changing the value of the control."""
        def event_handler(value):
            result = handler(value)
            if result is False:
                return js.Boolean(False)
        self.select.events.on('beforeChange', create_proxy(event_handler))

    def on_before_change_options(self, handler: Callable[[List[Dict[str, Any]]], Union[bool, None]]) -> None:
        """Fires before changing a list of Select options."""
        def event_handler(options):
            options_py = [option.to_py() for option in options]
            result = handler(options_py)
            if result is False:
                return js.Boolean(False)
        self.select.events.on('beforeChangeOptions', create_proxy(event_handler))

    def on_before_change_properties(self, handler: Callable[[Dict[str, Any]], Union[bool, None]]) -> None:
        """Fires before configuration attributes are changed dynamically."""
        def event_handler(properties):
            result = handler(properties.to_py())
            if result is False:
                return js.Boolean(False)
        self.select.events.on('beforeChangeProperties', create_proxy(event_handler))

    def on_before_hide(self, handler: Callable[[Union[str, int], bool], Union[bool, None]]) -> None:
        """Fires before the control is hidden."""
        def event_handler(value, init):
            result = handler(value, init)
            if result is False:
                return js.Boolean(False)
        self.select.events.on('beforeHide', create_proxy(event_handler))

    def on_before_show(self, handler: Callable[[Union[str, int]], Union[bool, None]]) -> None:
        """Fires before the control is shown."""
        def event_handler(value):
            result = handler(value)
            if result is False:
                return js.Boolean(False)
        self.select.events.on('beforeShow', create_proxy(event_handler))

    def on_before_validate(self, handler: Callable[[Union[str, int]], Union[bool, None]]) -> None:
        """Fires before the control value is validated."""
        def event_handler(value):
            result = handler(value)
            if result is False:
                return js.Boolean(False)
        self.select.events.on('beforeValidate', create_proxy(event_handler))

    def on_blur(self, handler: Callable[[Union[str, int]], None]) -> None:
        """Fires when the control has lost focus."""
        self.add_event_handler('blur', handler)

    def on_change(self, handler: Callable[[Union[str, int]], None]) -> None:
        """Fires on changing the value of the control."""
        self.add_event_handler('change', handler)

    def on_change_options(self, handler: Callable[[List[Dict[str, Any]]], None]) -> None:
        """Fires on changing a list of Select options."""
        self.add_event_handler('changeOptions', handler)

    def on_focus(self, handler: Callable[[Union[str, int]], None]) -> None:
        """Fires when the control has received focus."""
        self.add_event_handler('focus', handler)

    def on_keydown(self, handler: Callable[[Any], None]) -> None:
        """Fires when any key is pressed and the control is in focus."""
        self.add_event_handler('keydown', handler)



from typing import Union, Dict, Any, Callable, List

from pyodide.ffi import create_proxy


class ComboConfig:
    """
    Configuration class for the Combo control.
    """
    def __init__(self,
                 name: str = None,
                 id: str = None,
                 data: List[Dict[str, Any]] = None,
                 value: Union[str, int, List[Union[str, int]]] = None,
                 css: str = None,
                 disabled: bool = False,
                 height: Union[str, int] = "content",
                 hidden: bool = False,
                 padding: Union[str, int] = "5px",
                 required: bool = False,
                 validation: Callable[[Any], bool] = None,
                 width: Union[str, int] = "content",
                 filter: Callable[[Any], bool] = None,
                 eventHandlers: Dict[str, Callable] = None,
                 itemHeight: int = 32,
                 itemsCount: bool = False,
                 listHeight: int = 224,
                 multiselection: bool = False,
                 newOptions: bool = False,
                 placeholder: str = None,
                 readOnly: bool = False,
                 selectAllButton: bool = False,
                 template: str = None,
                 virtual: bool = False,
                 hiddenLabel: bool = False,
                 label: str = None,
                 labelPosition: str = "top",
                 labelWidth: Union[str, int] = None,
                 helpMessage: str = None,
                 preMessage: str = None,
                 successMessage: str = None,
                 errorMessage: str = None):
        """
        :param name: (Optional) The name of the control.
        :param id: (Optional) The id of the control.
        :param data: (Optional) An array of Combo options.
        :param value: (Optional) The id(s) of options to be selected.
        :param css: (Optional) Adds style classes to the control.
        :param disabled: (Optional) Whether the control is disabled.
        :param height: (Optional) The height of the control.
        :param hidden: (Optional) Whether the control is hidden.
        :param padding: (Optional) Sets padding between the cell and border.
        :param required: (Optional) Whether the control is required.
        :param validation: (Optional) A callback function for validation.
        :param width: (Optional) The width of the control.
        :param filter: (Optional) Custom function for filtering options.
        :param eventHandlers: (Optional) Event handlers for custom templates.
        :param itemHeight: (Optional) Height of a cell in the options list.
        :param itemsCount: (Optional) Shows the total number of selected options.
        :param listHeight: (Optional) Height of the list of options.
        :param multiselection: (Optional) Enables multiple selection.
        :param newOptions: (Optional) Allows users to add new options.
        :param placeholder: (Optional) Placeholder in the input field.
        :param readOnly: (Optional) Makes Combo readonly.
        :param selectAllButton: (Optional) Shows the Select All button.
        :param template: (Optional) Template for displaying options.
        :param virtual: (Optional) Enables dynamic loading of data.
        :param hiddenLabel: (Optional) Makes the label invisible.
        :param label: (Optional) Specifies a label for the control.
        :param labelPosition: (Optional) Position of the label.
        :param labelWidth: (Optional) Width of the label.
        :param helpMessage: (Optional) Adds a help message to the control.
        :param preMessage: (Optional) Instructions for interacting with the control.
        :param successMessage: (Optional) Message after successful validation.
        :param errorMessage: (Optional) Message after validation error.
        """
        self.type = "combo"
        self.name = name
        self.id = id
        self.data = data
        self.value = value
        self.css = css
        self.disabled = disabled
        self.height = height
        self.hidden = hidden
        self.padding = padding
        self.required = required
        self.validation = validation
        self.width = width
        self.filter = filter
        self.eventHandlers = eventHandlers
        self.itemHeight = itemHeight
        self.itemsCount = itemsCount
        self.listHeight = listHeight
        self.multiselection = multiselection
        self.newOptions = newOptions
        self.placeholder = placeholder
        self.readOnly = readOnly
        self.selectAllButton = selectAllButton
        self.template = template
        self.virtual = virtual
        self.hiddenLabel = hiddenLabel
        self.label = label
        self.labelPosition = labelPosition
        self.labelWidth = labelWidth
        self.helpMessage = helpMessage
        self.preMessage = preMessage
        self.successMessage = successMessage
        self.errorMessage = errorMessage

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the ComboConfig into a dictionary format.
        """
        config_dict = {
            'type': self.type,
            'name': self.name,
            'id': self.id,
            'data': self.data,
            'value': self.value,
            'css': self.css,
            'disabled': self.disabled,
            'height': self.height,
            'hidden': self.hidden,
            'padding': self.padding,
            'required': self.required,
            'width': self.width,
            'validation': self.validation,
            'filter': self.filter,
            'eventHandlers': self.eventHandlers,
            'itemHeight': self.itemHeight,
            'itemsCount': self.itemsCount,
            'listHeight': self.listHeight,
            'multiselection': self.multiselection,
            'newOptions': self.newOptions,
            'placeholder': self.placeholder,
            'readOnly': self.readOnly,
            'selectAllButton': self.selectAllButton,
            'template': self.template,
            'virtual': self.virtual,
            'hiddenLabel': self.hiddenLabel,
            'label': self.label,
            'labelPosition': self.labelPosition,
            'labelWidth': self.labelWidth,
            'helpMessage': self.helpMessage,
            'preMessage': self.preMessage,
            'successMessage': self.successMessage,
            'errorMessage': self.errorMessage
        }
        # Remove None values
        config_dict = {k: v for k, v in config_dict.items() if v is not None}

        # Handle functions
        if 'validation' in config_dict:
            config_dict['validation'] = create_proxy(self.validation)
        if 'filter' in config_dict:
            config_dict['filter'] = create_proxy(self.filter)
        if 'eventHandlers' in config_dict:
            event_handlers = {k: create_proxy(v) for k, v in config_dict['eventHandlers'].items()}
            config_dict['eventHandlers'] = event_handlers

        return config_dict



from typing import List, Union, Dict, Any


class TabConfig:
    """
    Configuration class for individual tabs in the Tabbar.
    """
    def __init__(self,
                 id: str,
                 tab: str = None,
                 tabCss: str = None,
                 css: str = None,
                 header: str = None,
                 html: str = None,
                 padding: Union[int, str] = None,
                 tabWidth: Union[int, str] = None,
                 tabHeight: Union[int, str] = None):
        """
        :param id: (Required) The id of the tab.
        :param tab: (Optional) The name of the tab.
        :param tabCss: (Optional) The CSS class used for the tab.
        :param css: (Optional) The CSS class used for the cell.
        :param header: (Optional) The header of the cell.
        :param html: (Optional) HTML content for the tab.
        :param padding: (Optional) The distance between the content and border.
        :param tabWidth: (Optional) The width of the tab.
        :param tabHeight: (Optional) The height of the tab.
        """
        self.id = id
        self.tab = tab
        self.tabCss = tabCss
        self.css = css
        self.header = header
        self.html = html
        self.padding = padding
        self.tabWidth = tabWidth
        self.tabHeight = tabHeight

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the TabConfig into a dictionary format.
        """
        config_dict = {
            'id': self.id,
            'tab': self.tab,
            'tabCss': self.tabCss,
            'css': self.css,
            'header': self.header,
            'html': self.html,
            'padding': self.padding,
            'tabWidth': self.tabWidth,
            'tabHeight': self.tabHeight
        }
        # Remove None values
        return {k: v for k, v in config_dict.items() if v is not None}


class TabbarConfig:
    """
    Configuration class for Tabbar.
    """
    def __init__(self,
                 views: List[TabConfig],
                 activeTab: str = None,
                 closable: Union[bool, List[str]] = None,
                 css: str = None,
                 disabled: Union[str, List[str]] = None,
                 mode: str = None,
                 noContent: bool = None,
                 tabAlign: str = None,
                 tabAutoHeight: bool = None,
                 tabAutoWidth: bool = None,
                 tabHeight: Union[int, str] = None,
                 tabWidth: Union[int, str] = None):
        """
        :param views: (Required) Defines the configuration of tabs.
        :param activeTab: (Optional) Sets the currently active tab.
        :param closable: (Optional) Adds close buttons for tabs.
        :param css: (Optional) The CSS class(es) applied to Tabbar.
        :param disabled: (Optional) Makes a tab or tabs disabled.
        :param mode: (Optional) Specifies the mode of displaying the tabbar.
        :param noContent: (Optional) Defines whether tabs contain any content.
        :param tabAlign: (Optional) Sets alignment for tabs.
        :param tabAutoHeight: (Optional) Adjusts tab height automatically.
        :param tabAutoWidth: (Optional) Adjusts tab width automatically.
        :param tabHeight: (Optional) Sets the height of a tab.
        :param tabWidth: (Optional) Sets the width of a tab.
        """
        self.views = views
        self.activeTab = activeTab
        self.closable = closable
        self.css = css
        self.disabled = disabled
        self.mode = mode
        self.noContent = noContent
        self.tabAlign = tabAlign
        self.tabAutoHeight = tabAutoHeight
        self.tabAutoWidth = tabAutoWidth
        self.tabHeight = tabHeight
        self.tabWidth = tabWidth

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the TabbarConfig into a dictionary format.
        """
        config_dict = {
            'views': [view.to_dict() for view in self.views],
            'activeTab': self.activeTab,
            'closable': self.closable,
            'css': self.css,
            'disabled': self.disabled,
            'mode': self.mode,
            'noContent': self.noContent,
            'tabAlign': self.tabAlign,
            'tabAutoHeight': self.tabAutoHeight,
            'tabAutoWidth': self.tabAutoWidth,
            'tabHeight': self.tabHeight,
            'tabWidth': self.tabWidth
        }
        # Remove None values
        return {k: v for k, v in config_dict.items() if v is not None}



"""
pyTincture Colorpicker widget implementation
"""

from typing import Any, Callable, Dict, List, Union
import json
from pyodide.ffi import create_proxy
import js

from .colorpicker_config import ColorpickerConfig

class Colorpicker:
    def __init__(self, config: ColorpickerConfig = None, widget_parent: Any = None):
        """Initializes the Colorpicker instance."""
        if config is None:
            config = ColorpickerConfig()
        self.colorpicker = js.dhx.Colorpicker.new(widget_parent, js.JSON.parse(json.dumps(config.to_dict())))
    
    """ Colorpicker API Functions """

    def clear(self) -> None:
        """Clears the value set in the colorpicker and removes focus."""
        self.colorpicker.clear()

    def destructor(self) -> None:
        """Releases the occupied resources."""
        self.colorpicker.destructor()

    def get_current_mode(self) -> str:
        """Returns the current mode of displaying Colorpicker ('palette' or 'picker')."""
        return self.colorpicker.getCurrentMode()

    def get_custom_colors(self) -> List[str]:
        """Returns an array of selected custom colors."""
        return list(self.colorpicker.getCustomColors())

    def get_value(self) -> str:
        """Returns the code of a selected color in the Hex format."""
        return self.colorpicker.getValue()

    def paint(self) -> None:
        """Repaints Colorpicker on a page."""
        self.colorpicker.paint()

    def set_current_mode(self, view: str) -> None:
        """Shows Colorpicker either in the 'palette' or 'picker' mode."""
        self.colorpicker.setCurrentMode(view)

    def set_custom_colors(self, custom_colors: List[str]) -> None:
        """Sets custom colors that will be displayed at the bottom of the palette."""
        self.colorpicker.setCustomColors(custom_colors)

    def set_focus(self, value: str) -> None:
        """Sets focus on the specified color value."""
        self.colorpicker.setFocus(value)

    def set_value(self, value: str) -> None:
        """Selects a color in Colorpicker."""
        self.colorpicker.setValue(value)

    """ Colorpicker Events """

    def add_event_handler(self, event_name: str, handler: Callable) -> None:
        """Helper to add event handlers dynamically."""
        event_proxy = create_proxy(handler)
        self.colorpicker.events.on(event_name, event_proxy)

    def on_apply(self, handler: Callable[[], None]) -> None:
        """Fires on clicking the 'Select' button."""
        self.add_event_handler('apply', handler)

    def on_before_change(self, handler: Callable[[str], Union[bool, None]]) -> None:
        """Fires before changing the selected color in Colorpicker."""
        def event_handler(color):
            result = handler(color)
            if result is False:
                return js.Boolean(False)
            # Do not return anything to allow the change
        event_proxy = create_proxy(event_handler)
        self.colorpicker.events.on('beforeChange', event_proxy)

    def on_cancel_click(self, handler: Callable[[], None]) -> None:
        """Fires on clicking the 'Cancel' button."""
        self.add_event_handler('cancelClick', handler)

    def on_change(self, handler: Callable[[str], None]) -> None:
        """Fires on changing the selected color in Colorpicker."""
        self.add_event_handler('change', handler)

    def on_mode_change(self, handler: Callable[[str], None]) -> None:
        """Fires on changing the mode of the colorpicker."""
        self.add_event_handler('modeChange', handler)

    """ Colorpicker Properties """

    @property
    def css(self) -> str:
        """Gets or sets custom CSS classes for Colorpicker."""
        return self.colorpicker.config.css

    @css.setter
    def css(self, value: str) -> None:
        self.colorpicker.config.css = value

    @property
    def custom_colors(self) -> List[str]:
        """Gets or sets custom colors displayed at the bottom of Colorpicker."""
        return self.colorpicker.config.customColors

    @custom_colors.setter
    def custom_colors(self, value: List[str]) -> None:
        self.colorpicker.config.customColors = value

    @property
    def gray_shades(self) -> bool:
        """Gets or sets whether the gray shades are displayed in the palette."""
        return self.colorpicker.config.grayShades

    @gray_shades.setter
    def gray_shades(self, value: bool) -> None:
        self.colorpicker.config.grayShades = value

    @property
    def mode(self) -> str:
        """Gets or sets the mode of displaying the colorpicker ('palette' or 'picker')."""
        return self.colorpicker.config.mode

    @mode.setter
    def mode(self, value: str) -> None:
        self.colorpicker.config.mode = value

    @property
    def palette(self) -> List[List[str]]:
        """Gets or sets the arrays of colors shown in the colorpicker."""
        return self.colorpicker.config.palette

    @palette.setter
    def palette(self, value: List[List[str]]) -> None:
        self.colorpicker.config.palette = value

    @property
    def palette_only(self) -> bool:
        """Gets or sets whether Colorpicker is shown only in the palette mode."""
        return self.colorpicker.config.paletteOnly

    @palette_only.setter
    def palette_only(self, value: bool) -> None:
        self.colorpicker.config.paletteOnly = value

    @property
    def picker_only(self) -> bool:
        """Gets or sets whether Colorpicker is shown only in the picker mode."""
        return self.colorpicker.config.pickerOnly

    @picker_only.setter
    def picker_only(self, value: bool) -> None:
        self.colorpicker.config.pickerOnly = value

    @property
    def transparency(self) -> bool:
        """Gets or sets whether the transparency scale is displayed in the picker mode."""
        return self.colorpicker.config.transparency

    @transparency.setter
    def transparency(self, value: bool) -> None:
        self.colorpicker.config.transparency = value

    @property
    def width(self) -> Union[str, int]:
        """Gets or sets the width of Colorpicker."""
        return self.colorpicker.config.width

    @width.setter
    def width(self, value: Union[str, int]) -> None:
        self.colorpicker.config.width = value



import sys
from dhxpyt.layout import MainWindow, Layout, LayoutConfig, CellConfig
from dhxpyt.toolbar import ButtonConfig, ToolbarConfig
from dhxpyt.sidebar import NavItemConfig, SidebarConfig
from dhxpyt.grid import GridConfig, GridColumnConfig
from pyodide.ffi import create_proxy

# Sample data for the grid
SAMPLE_DATA = [
    {"id": 1, "name": "Item 1", "value": 100},
    {"id": 2, "name": "Item 2", "value": 200},
    {"id": 3, "name": "Item 3", "value": 300}
]

class QuickstartMain(Layout):
    layout_config = LayoutConfig(
        type="line",
        cols=[
            CellConfig(id="sidebar", width="auto"),
            CellConfig(id="content")
        ]
    )

    def load_ui(self):
        # Sidebar configuration
        sidebar_items = [
            NavItemConfig(id="hamburger", icon="mdi mdi-menu"),
            NavItemConfig(id="items", value="Items", icon="mdi mdi-view-list")
        ]
        sidebar_config = SidebarConfig(data=sidebar_items, collapsed=False)
        self.sidebar = self.add_sidebar(id="sidebar", sidebar_config=sidebar_config)
        self.sidebar.on_click(create_proxy(self.handle_sidebar_click))

        # Content layout with toolbar and grid
        content_layout_config = LayoutConfig(
            type="line",
            rows=[
                CellConfig(id="toolbar", height="auto"),
                CellConfig(id="grid")
            ]
        )
        self.content_layout = self.add_layout("content", content_layout_config)

        # Toolbar configuration
        toolbar_config = ToolbarConfig(data=[
            ButtonConfig(id="add", value="Add Item", icon="mdi mdi-plus"),
            ButtonConfig(id="refresh", value="Refresh", icon="mdi mdi-refresh")
        ])
        self.toolbar = self.content_layout.add_toolbar(id="toolbar", toolbar_config=toolbar_config)

        # Grid configuration
        grid_columns = [
            GridColumnConfig(id="id", width=100, header=[{"text": "ID"}]),
            GridColumnConfig(id="name", width=200, header=[{"text": "Name"}]),
            GridColumnConfig(id="value", width=150, header=[{"text": "Value"}])
        ]
        grid_config = GridConfig(columns=grid_columns, data=SAMPLE_DATA)
        self.grid = self.content_layout.add_grid(id="grid", grid_config=grid_config)

    def handle_sidebar_click(self, id, event):
        if id == "hamburger":
            self.sidebar.toggle()

class QuickstartApp(MainWindow):
    def load_ui(self):
        self.set_theme("dark")
        self.main_layout = QuickstartMain(parent=self)
        self.attach("mainwindow", self.main_layout.layout)

if __name__ == "__main__" and sys.platform != "emscripten":
    from pytincture import launch_service
    launch_service()



from .form import Form
from .form_config import FormConfig
from .controls import Avatar, Button, Checkbox, CheckboxGroup, Colorpicker, Combo, Container, Datepicker, Fieldset, Input, RadioGroup, Select, SimpleVault, Slider, Spacer, Text, Textarea, Timepicker, Toggle, ToggleGroup
from .controls import AvatarConfig, ButtonConfig, CheckboxConfig, CheckboxGroupConfig, ColorpickerConfig, ComboConfig, ContainerConfig, DatepickerConfig, FieldsetConfig, InputConfig, RadioGroupConfig, SelectConfig, SimpleVaultConfig, SliderConfig, SpacerConfig, TextConfig, TextareaConfig, TimepickerConfig, ToggleConfig, ToggleGroupConfig


__all__ = ["Form", "FormConfig", "Avatar", "Button", "Checkbox", "CheckboxGroup", "Colorpicker", "Combo", "Container", "Datepicker", "Fieldset", "Input", "RadioGroup", "Select", "SimpleVault", "Slider", "Spacer", "Text", "Textarea", "Timepicker", "Toggle", "ToggleGroup", "AvatarConfig", "ButtonConfig", "CheckboxConfig", "CheckboxGroupConfig", "ColorpickerConfig", "ComboConfig", "ContainerConfig", "DatepickerConfig", "FieldsetConfig", "InputConfig", "RadioGroupConfig", "SelectConfig", "SimpleVaultConfig", "SliderConfig", "SpacerConfig", "TextConfig", "TextareaConfig", "TimepickerConfig", "ToggleConfig", "ToggleGroupConfig"]


"""
ToggleGroup control implementation for the Form widget
"""

from typing import Any, Callable, Dict, Union, List
import json
from pyodide.ffi import create_proxy
import js

from .togglegroup_config import ToggleGroupConfig


class ToggleGroup:
    def __init__(self, config: ToggleGroupConfig = None, widget_parent: Any = None):
        """Initializes the ToggleGroup control."""
        if config is None:
            config = ToggleGroupConfig()
        config_dict = config.to_dict()
        self.togglegroup = js.dhx.FormControl.new(widget_parent, js.JSON.parse(json.dumps(config_dict)))

    """ ToggleGroup API Functions """

    def blur(self) -> None:
        """Removes focus from the ToggleGroup control."""
        self.togglegroup.blur()

    def destructor(self) -> None:
        """Removes the ToggleGroup instance and releases the occupied resources."""
        self.togglegroup.destructor()

    def disable(self, id: str = None) -> None:
        """Disables the ToggleGroup control or a specific element."""
        self.togglegroup.disable(id)

    def enable(self, id: str = None) -> None:
        """Enables the ToggleGroup control or a specific element."""
        self.togglegroup.enable(id)

    def focus(self, id: str = None) -> None:
        """Sets focus to an option of the ToggleGroup control by its id."""
        self.togglegroup.focus(id)

    def get_properties(self, id: str = None) -> Dict[str, Any]:
        """Returns the configuration attributes of the control or its toggle."""
        props = self.togglegroup.getProperties(id)
        return props.to_py() if props else {}

    def get_value(self, id: str = None) -> Union[str, int, bool, Dict[str, Union[str, int, bool]]]:
        """Returns the current value/state of a toggle(s)."""
        return self.togglegroup.getValue(id).to_py()

    def hide(self, id: str = None) -> None:
        """Hides either an option of ToggleGroup or the whole ToggleGroup."""
        self.togglegroup.hide(id)

    def is_disabled(self, id: str = None) -> bool:
        """Checks whether the ToggleGroup control or a specific element is disabled."""
        return self.togglegroup.isDisabled(id)

    def is_selected(self, id: str = None) -> Union[bool, Dict[str, bool]]:
        """Checks whether a toggle of the ToggleGroup control is selected."""
        result = self.togglegroup.isSelected(id)
        return result.to_py() if isinstance(result, js.Object) else result

    def is_visible(self, id: str = None) -> bool:
        """Checks whether the ToggleGroup control or a specific element is visible."""
        return self.togglegroup.isVisible(id)

    def set_properties(self, config: Dict[str, Any], id: str = None) -> None:
        """Allows changing configuration attributes dynamically."""
        self.togglegroup.setProperties(js.JSON.parse(json.dumps(config)), id)

    def set_value(self, value: Dict[str, bool]) -> None:
        """Defines the state of the option's elements."""
        self.togglegroup.setValue(js.JSON.parse(json.dumps(value)))

    def show(self, id: str = None) -> None:
        """Shows either an option of ToggleGroup or the whole ToggleGroup."""
        self.togglegroup.show(id)

    """ ToggleGroup Events """

    def add_event_handler(self, event_name: str, handler: Callable) -> None:
        """Helper to add event handlers dynamically."""
        event_proxy = create_proxy(handler)
        self.togglegroup.events.on(event_name, event_proxy)

    def on_after_change_properties(self, handler: Callable[[Dict[str, Any], str], None]) -> None:
        """Fires after configuration attributes have been changed dynamically."""
        self.add_event_handler('afterChangeProperties', handler)

    def on_after_hide(self, handler: Callable[[Dict[str, Any], str, bool], None]) -> None:
        """Fires after the control or its toggle is hidden."""
        self.add_event_handler('afterHide', handler)

    def on_after_show(self, handler: Callable[[Dict[str, Any], str], None]) -> None:
        """Fires after the control or its toggle is shown."""
        self.add_event_handler('afterShow', handler)

    def on_before_change(self, handler: Callable[[Dict[str, Any]], Union[bool, None]]) -> None:
        """Fires before changing the value of the control."""
        def event_handler(value):
            result = handler(value.to_py())
            if result is False:
                return js.Boolean(False)
        self.togglegroup.events.on('beforeChange', create_proxy(event_handler))

    def on_before_change_properties(self, handler: Callable[[Dict[str, Any], str], Union[bool, None]]) -> None:
        """Fires before configuration attributes are changed dynamically."""
        def event_handler(properties, id=None):
            result = handler(properties.to_py(), id)
            if result is False:
                return js.Boolean(False)
        self.togglegroup.events.on('beforeChangeProperties', create_proxy(event_handler))

    def on_before_hide(self, handler: Callable[[Dict[str, Any], str, bool], Union[bool, None]]) -> None:
        """Fires before the control or its toggle is hidden."""
        def event_handler(value, id=None, init=False):
            result = handler(value.to_py(), id, init)
            if result is False:
                return js.Boolean(False)
        self.togglegroup.events.on('beforeHide', create_proxy(event_handler))

    def on_before_show(self, handler: Callable[[Dict[str, Any], str], Union[bool, None]]) -> None:
        """Fires before the control or its toggle is shown."""
        def event_handler(value, id=None):
            result = handler(value.to_py(), id)
            if result is False:
                return js.Boolean(False)
        self.togglegroup.events.on('beforeShow', create_proxy(event_handler))

    def on_blur(self, handler: Callable[[Dict[str, Any], str], None]) -> None:
        """Fires when the ToggleGroup control has lost focus."""
        self.add_event_handler('blur', handler)

    def on_change(self, handler: Callable[[Dict[str, Any]], None]) -> None:
        """Fires on changing the value of the control."""
        self.add_event_handler('change', handler)

    def on_focus(self, handler: Callable[[Dict[str, Any], str], None]) -> None:
        """Fires when the ToggleGroup control has received focus."""
        self.add_event_handler('focus', handler)

    def on_keydown(self, handler: Callable[[Any, str], None]) -> None:
        """Fires when any key is pressed and a toggle is in focus."""
        self.add_event_handler('keydown', handler)
